#!/usr/bin/env python3
#
# Tests the MathML expression writer in presentation mode.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import re
import unittest

import myokit
import myokit.formats.mathml as mathml


class PresentationMathMLTest(unittest.TestCase):
    """
    Tests export of presentation MathML.
    """

    @classmethod
    def setUpClass(cls):
        model = myokit.Model()
        component = model.add_component('c')
        cls.avar = component.add_variable('a')

        cls.w = mathml.MathMLExpressionWriter()
        cls.w.set_mode(presentation=True)

        # MathML opening and closing tags
        cls._math = re.compile(r'^<math [^>]+>(.*)</math>$', re.S)

    def assertWrite(self, expression, xml):
        """ Assert writing an ``expression`` results in the given ``xml``. """
        x = self.w.ex(expression)
        m = self._math.match(x)
        self.assertTrue(m)
        self.assertEqual(m.group(1), xml)

    def test_arithmetic_binary(self):
        # Tests writing basic arithmetic operators

        a = myokit.Name(self.avar)
        b = myokit.Number('12', 'pF')
        ca = '<mi>c.a</mi>'
        cb = '<mn>12.0</mn>'

        # Plus
        x = myokit.Plus(a, b)
        self.assertWrite(x, '<mrow>' + ca + '<mo>+</mo>' + cb + '</mrow>')

        # Minus
        x = myokit.Minus(a, b)
        self.assertWrite(x, '<mrow>' + ca + '<mo>-</mo>' + cb + '</mrow>')

        # Multiply
        x = myokit.Multiply(a, b)
        self.assertWrite(x, '<mrow>' + ca + '<mo>*</mo>' + cb + '</mrow>')

        # Divide
        x = myokit.Divide(a, b)
        self.assertWrite(x, '<mfrac>' + ca + cb + '</mfrac>')

    def test_arithmetic_unary(self):
        # Tests writing prefix operators

        b = myokit.Number('12', 'pF')
        cb = '<mn>12.0</mn>'

        # Prefix plus
        x = myokit.PrefixPlus(b)
        self.assertWrite(x, '<mrow><mo>+</mo>' + cb + '</mrow>')

        # Prefix minus
        x = myokit.PrefixMinus(b)
        self.assertWrite(x, '<mrow><mo>-</mo>' + cb + '</mrow>')

    def test_conditionals(self):
        # Tests if and piecewise writing

        a = myokit.Name('a')
        b = myokit.Number(1)
        c = myokit.Number(1)
        cond1 = myokit.parse_expression('5 > 3')
        cond2 = myokit.parse_expression('2 < 1')

        ca = '<mi>a</mi>'
        cb = '<mn>1.0</mn>'
        cc = '<mn>1.0</mn>'
        c1 = '<mrow><mn>5.0</mn><mo>&gt;</mo><mn>3.0</mn></mrow>'
        c2 = '<mrow><mn>2.0</mn><mo>&lt;</mo><mn>1.0</mn></mrow>'

        # If
        x = myokit.If(cond1, a, b)
        self.assertWrite(
            x,
            '<piecewise>'
            '<piece>' + ca + c1 + '</piece>'
            '<otherwise>' + cb + '</otherwise>'
            '</piecewise>'
        )
        # Piecewise
        x = myokit.Piecewise(cond1, a, cond2, b, c)
        self.assertWrite(
            x,
            '<piecewise>'
            '<piece>' + ca + c1 + '</piece>'
            '<piece>' + cb + c2 + '</piece>'
            '<otherwise>' + cc + '</otherwise>'
            '</piecewise>'
        )

    def test_creation(self):
        # Test creation via formats.ewriter

        w = myokit.formats.ewriter('mathml')
        self.assertIsInstance(self.w, mathml.MathMLExpressionWriter)

    def test_functions(self):
        # Tests writing basic functions

        a = myokit.Name(self.avar)
        b = myokit.Number('12', 'pF')
        ca = '<mi>c.a</mi>'
        cb = '<mn>12.0</mn>'

        # Power
        x = myokit.Power(a, b)
        self.assertWrite(x, '<msup>' + ca + cb + '</msup>')

        # Sqrt
        x = myokit.Sqrt(b)
        self.assertWrite(
            x, '<mrow><mi>root</mi><mfenced>' + cb + '</mfenced></mrow>')

        # Exp
        x = myokit.Exp(a)
        self.assertWrite(x, '<msup><mi>e</mi>' + ca + '</msup>')

        # Log(a)
        x = myokit.Log(b)
        self.assertWrite(
            x, '<mrow><mi>ln</mi><mfenced>' + cb + '</mfenced></mrow>')

        # Log(a, b)
        x = myokit.Log(a, b)
        self.assertWrite(
            x,
            '<mrow><msub><mi>log</mi>' + cb + '</msub>'
            '<mfenced>' + ca + '</mfenced></mrow>'
        )

        # Log10
        x = myokit.Log10(b)
        self.assertWrite(
            x, '<mrow><mi>log</mi><mfenced>' + cb + '</mfenced></mrow>')

        # Floor
        x = myokit.Floor(b)
        self.assertWrite(
            x, '<mrow><mi>floor</mi><mfenced>' + cb + '</mfenced></mrow>')

        # Ceil
        x = myokit.Ceil(b)
        self.assertWrite(
            x, '<mrow><mi>ceiling</mi><mfenced>' + cb + '</mfenced></mrow>')

        # Abs
        x = myokit.Abs(b)
        self.assertWrite(
            x, '<mrow><mi>abs</mi><mfenced>' + cb + '</mfenced></mrow>')

        # Quotient
        x = myokit.Quotient(a, b)
        self.assertWrite(x, '<mrow>' + ca + '<mo>//</mo>' + cb + '</mrow>')

        # Remainder
        x = myokit.Remainder(a, b)
        self.assertWrite(x, '<mrow>' + ca + '<mo>%</mo>' + cb + '</mrow>')

    def test_inequalities(self):
        # Test writing (in)equalities

        a = myokit.Name(self.avar)
        b = myokit.Number('12', 'pF')
        ca = '<mi>c.a</mi>'
        cb = '<mn>12.0</mn>'

        # Equal
        x = myokit.Equal(a, b)
        self.assertWrite(x, '<mrow>' + ca + '<mo>==</mo>' + cb + '</mrow>')

        # NotEqual
        x = myokit.NotEqual(a, b)
        self.assertWrite(x, '<mrow>' + ca + '<mo>!=</mo>' + cb + '</mrow>')

        # More
        x = myokit.More(a, b)
        self.assertWrite(x, '<mrow>' + ca + '<mo>&gt;</mo>' + cb + '</mrow>')

        # Less
        x = myokit.Less(a, b)
        self.assertWrite(x, '<mrow>' + ca + '<mo>&lt;</mo>' + cb + '</mrow>')

        # MoreEqual
        # Named version &ge; is not output, shows decimal code instead
        x = myokit.MoreEqual(a, b)
        self.assertWrite(
            x, '<mrow>' + ca + '<mo>&#8805;</mo>' + cb + '</mrow>')

        # LessEqual
        # Named version &le; is not output, shows decimal code instead
        x = myokit.LessEqual(a, b)
        self.assertWrite(
            x, '<mrow>' + ca + '<mo>&#8804;</mo>' + cb + '</mrow>')

    def test_logic_operators(self):
        # Tests writing logic operators

        cond1 = myokit.parse_expression('5 > 3')
        cond2 = myokit.parse_expression('2 < 1')
        c1 = '<mrow><mn>5.0</mn><mo>&gt;</mo><mn>3.0</mn></mrow>'
        c2 = '<mrow><mn>2.0</mn><mo>&lt;</mo><mn>1.0</mn></mrow>'

        # Not
        x = myokit.Not(cond1)
        self.assertWrite(
            x, '<mrow><mo>(</mo><mo>not</mo>' + c1 + '<mo>)</mo></mrow>')

        # And
        x = myokit.And(cond1, cond2)
        self.assertWrite(x, '<mrow>' + c1 + '<mo>and</mo>' + c2 + '</mrow>')

        # Or
        x = myokit.Or(cond1, cond2)
        self.assertWrite(x, '<mrow>' + c1 + '<mo>or</mo>' + c2 + '</mrow>')

    def test_name_and_numbers(self):
        # Test name and number writing

        # Name
        self.assertWrite(myokit.Name(self.avar), '<mi>c.a</mi>')

        # Number with unit
        self.assertWrite(myokit.Number(12, 'pF'), '<mn>12.0</mn>')

        # Number without unit
        self.assertWrite(myokit.Number(1), '<mn>1.0</mn>')

        # E-notation is allowed in presentation MathML
        self.assertWrite(myokit.Number(1e-6), '<mn>1e-06</mn>')
        self.assertWrite(myokit.Number(1e24), '<mn>1e+24</mn>')

    def test_trig_basic(self):
        # Test writing basic trig functions

        b = myokit.Number('12', 'pF')
        cb = '<mn>12.0</mn>'

        # Sin
        x = myokit.Sin(b)
        self.assertWrite(
            x, '<mrow><mi>sin</mi><mfenced>' + cb + '</mfenced></mrow>')

        # Cos
        x = myokit.Cos(b)
        self.assertWrite(
            x, '<mrow><mi>cos</mi><mfenced>' + cb + '</mfenced></mrow>')

        # Tan
        x = myokit.Tan(b)
        self.assertWrite(
            x, '<mrow><mi>tan</mi><mfenced>' + cb + '</mfenced></mrow>')

        # ASin
        x = myokit.ASin(b)
        self.assertWrite(
            x, '<mrow><mi>arcsin</mi><mfenced>' + cb + '</mfenced></mrow>')

        # ACos
        x = myokit.ACos(b)
        self.assertWrite(
            x, '<mrow><mi>arccos</mi><mfenced>' + cb + '</mfenced></mrow>')

        # ATan
        x = myokit.ATan(b)
        self.assertWrite(
            x, '<mrow><mi>arctan</mi><mfenced>' + cb + '</mfenced></mrow>')

    def test_unknown_expression(self):
        # Test without a Myokit expression

        # Test without a Myokit expression
        self.assertRaisesRegex(
            ValueError, 'Unknown expression type', self.w.ex, 7)


if __name__ == '__main__':
    unittest.main()



#
# OpenCL information class
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import configparser
import os

import myokit

# Settings file
SETTINGS_FILE = os.path.join(myokit.DIR_USER, 'preferred-opencl-device.ini')

# Location of C source for OpenCL info module
SOURCE_FILE = 'opencl.c'


class OpenCL(myokit.CModule):
    """
    Tests for OpenCL support and can return information about opencl
    simulations.
    """
    # Unique id for this object
    _index = 0

    # Cached back-end object if compiled, False if compilation failed
    _instance = None

    # Cached compilation error message
    _message = None

    def __init__(self):
        super().__init__()
        # Create back-end and cache it
        OpenCL._index += 1
        mname = 'myokit_opencl_info_' + str(OpenCL._index)
        mname += '_' + str(myokit.pid_hash())
        fname = os.path.join(myokit.DIR_CFUNC, SOURCE_FILE)
        args = {'module_name': mname}

        # Define libraries
        libs = []
        flags = []
        import platform
        if platform.system() != 'Darwin':   # pragma: no macos cover
            libs.append('OpenCL')
        else:                               # pragma: no cover
            flags.append('-framework')
            flags.append('OpenCL')

        # Add include / linker paths
        libd = list(myokit.OPENCL_LIB)
        incd = list(myokit.OPENCL_INC)
        incd.append(myokit.DIR_CFUNC)

        try:
            OpenCL._message = None
            OpenCL._instance = self._compile(
                mname, fname, args, libs, libd, incd, larg=flags)
        except myokit.CompilationError as e:    # pragma: no cover
            OpenCL._instance = False
            OpenCL._message = str(e)

    @staticmethod
    def _get_instance():
        """
        Returns a cached back-end, creates and returns a new back-end or raises
        a :class:`NoOpenCLError`.
        """
        # No instance? Create it
        if OpenCL._instance is None:
            OpenCL()
        # Instance creation failed, raise exception
        if OpenCL._instance is False:   # pragma: no cover
            raise NoOpenCLError(
                'OpenCL support not found.\n' + OpenCL._message)
        # Return instance
        return OpenCL._instance

    @staticmethod
    def available():
        """
        Returns ``True`` if OpenCL support has been detected on this sytem and
        at least one platform and device were detected.
        """
        try:
            info = OpenCL.info()
        except Exception:   # pragma: no cover
            return False
        for p in info.platforms:
            for d in p.devices:
                return True
        return False    # pragma: no cover

    @staticmethod
    def current_info(formatted=False):
        """
        Returns a :class:`myokit.OpenCLPlatformInfo` object for the platform
        and device selected by the user, or chosen as default.

        If ``formatted=True`` is set, a formatted version of the information is
        returned instead.
        """
        platform, device = OpenCL.load_selection_bytes()
        info = OpenCL._get_instance().current(platform, device)
        info = OpenCLPlatformInfo(info)
        return info.format() if formatted else info

    @staticmethod
    def info(formatted=False):
        """
        Queries the OpenCL installation for the available platforms and
        devices and returns a :class:`myokit.OpenCLInfo` object.

        If ``formatted=True`` is set, a formatted version of the information is
        returned instead.
        """
        if not OpenCL.supported():  # pragma: no cover
            if formatted:
                return 'No OpenCL support detected.'
            return OpenCLInfo()

        info = OpenCLInfo(OpenCL._get_instance().info())
        if formatted:
            if len(info.platforms) == 0:  # pragma: no cover
                return 'OpenCL drivers detected, but no devices found.'
            return info.format()
        return info

    @staticmethod
    def load_selection():
        """
        Loads a platform/device selection from disk and returns a tuple
        ``(platform, device)``. Each entry in the tuple is either a string
        with the platform/device name, or ``None`` if no preference was set.
        """
        platform, device = OpenCL.load_selection_bytes()
        if platform is not None:
            platform = platform.decode('ascii')
        if device is not None:
            device = device.decode('ascii')

        return platform, device

    @staticmethod
    def load_selection_bytes():
        """
        Loads a platform/device selection from disk and returns a tuple
        ``(platform, device)``. Each entry in the tuple is either a string
        with the platform/device name, or ``None`` if no preference was set.
        """
        platform = device = None

        # Read ini file
        inifile = os.path.expanduser(SETTINGS_FILE)
        if os.path.isfile(inifile):
            config = configparser.RawConfigParser()
            config.read(inifile, encoding='ascii')

            def get(section, option):
                x = None
                if config.has_section(section):
                    if config.has_option(section, option):
                        x = config.get(section, option).strip()
                        if x:
                            return x
                return None

            platform = get('selection', 'platform')
            device = get('selection', 'device')

        # Ensure platform and device are ascii compatible byte strings, or None
        if platform is not None:
            platform = platform.encode('ascii')
        if device is not None:
            device = device.encode('ascii')

        return platform, device

    @staticmethod
    def save_selection(platform=None, device=None):
        """
        Stores a platform/device selection to disk.

        Both platform and device are identified by their names.
        """
        # Make sure platform and device can be stored as ascii
        if platform:
            platform = platform.encode('ascii').decode('ascii')
        if device:
            device = device.encode('ascii').decode('ascii')

        # Create configuration
        config = configparser.RawConfigParser()
        config.add_section('selection')
        if platform:
            config.set('selection', 'platform', platform)
        if device:
            config.set('selection', 'device', device)

        # Write configuration to ini file
        inifile = os.path.expanduser(SETTINGS_FILE)
        with open(inifile, 'w') as configfile:
            config.write(configfile)

    @staticmethod
    def selection_info():
        """
        Returns a list of platform/device combinations along with information
        allowing the user to select one.

        The returned output is a list of tuples, where each tuple has the form
        ``(platform_name, device_name, specs)``.

        A preferred device can be selected by passing one of the returned
        ``platform_name, device_name`` combinations to
        :meth:`OpenCL.set_preferred_device`.
        """
        devices = []
        for platform in OpenCL.info().platforms:
            for device in platform.devices:
                specs = clockspeed(device.clock)
                specs += ', ' + bytesize(device.globl) + ' global'
                specs += ', ' + bytesize(device.local) + ' local'
                specs += ', ' + bytesize(device.const) + ' const'
                devices.append((platform.name, device.name, specs))
        return devices

    @staticmethod
    def supported():
        """
        Returns ``True`` if OpenCL support has been detected on this system.
        """
        try:
            OpenCL._get_instance()
            return True
        except NoOpenCLError:       # pragma: no cover
            return False

    '''
    @staticmethod
    def test_build(code):
        """
        Tries building a kernel program on the currently selected platform and
        device and returns the compiler output.
        """
        try:
            cl = OpenCL._get_instance()
        except NoOpenCLError as e:
            return False

        # Get preferred platform/device combo from configuration file
        platform, device = myokit.OpenCL.load_selection_bytes()

        # Build and return
        return cl.build(platform, device, code)
    '''


class OpenCLInfo:
    """
    Represents information about the available OpenCL platforms and devices.

    Each ``OpenCLInfo`` object has a property ``platforms``, containing a tuple
    of :class:`OpenCLPlatformInfo` objects.

    ``OpenCLInfo`` objects are created and returned by :class:`myokit.OpenCL`.
    """
    def __init__(self, mcl_info=[]):
        # mcl_info is a python object returned by mcl_device_info (mcl.h)
        self.platforms = tuple([OpenCLPlatformInfo(x) for x in mcl_info])

    def format(self):
        """
        Returns a formatted string version of this object's information.
        """
        selected_platform, selected_device = myokit.OpenCL.load_selection()
        b = []
        for i, platform in enumerate(self.platforms):
            pname = 'Platform ' + str(i)
            if platform.name == selected_platform:  # pragma: no cover
                pname += '  <-- Selected platform in myokit.ini'
            b.append(pname)
            platform._format(b, selected_platform, selected_device, pre=' ')
        return '\n'.join(b)


class OpenCLPlatformInfo:
    """
    Represents information about an OpenCL platform.

    An ``OpenCLPlatformInfo`` object has the following properties:

    ``name`` (string)
        This platform's name.
    ``vendor`` (string)
        The vendor of this platform.
    ``version`` (string)
        The OpenCL version supported by this platform.
    ``profile`` (string)
        The supported OpenCL profile of this platform.
    ``extensions`` (string)
        The available OpenCL extensions on this platform.
    ``devices`` (tuple)
        A tuple of device information objects for the devices available on
        this platform. This field may be ``None``, in which case ``device``
        will be set instead.
    ``device`` (OpenCLDeviceInfo)
        An information objects for the device selected by the user, or chosen
        as the default device. This field may be ``None``, in which case
        ``devices`` will be set instead.

    ``OpenCLPlatformInfo`` objects are created as part of a :class:`OpenCLInfo`
    objects, as returned by most OpenCL enabled parts of Myokit.
    """
    def __init__(self, platform):
        self.name = platform['name'].strip()
        self.vendor = platform['vendor'].strip()
        self.version = platform['version'].strip()
        self.profile = platform['profile'].strip()
        self.extensions = platform['extensions'].strip().split()
        self.devices = platform.get('devices', None)
        if self.devices is not None:
            self.devices = tuple([OpenCLDeviceInfo(x) for x in self.devices])
        self.device = platform.get('device', None)
        if self.device is not None:
            self.device = OpenCLDeviceInfo(self.device)

    def format(self):
        """
        Returns a formatted string version of this object's information.
        """
        b = []
        b.append('Platform: ' + self.name)
        self._format(b, pre=' ', name=False)
        return '\n'.join(b)

    def _format(self, b, p_selected=None, d_selected=None, pre='', name=True):
        """
        Formats the information in this object and adds it to the list ``b``.
        """
        if name:
            b.append(pre + 'Name       : ' + self.name)
        b.append(pre + 'Vendor     : ' + self.vendor)
        b.append(pre + 'Version    : ' + self.version)
        b.append(pre + 'Profile    : ' + self.profile)
        b.append(pre + 'Extensions : ' + ' '.join(self.extensions))
        if self.devices is not None:
            b.append(pre + 'Devices:')
            for j, device in enumerate(self.devices):
                dname = pre + ' Device ' + str(j)
                if self.name == p_selected and device.name == d_selected:  # pragma: no cover # noqa
                    dname += '  <-- Selected device in myokit.ini'
                b.append(dname)
                device._format(b, pre + '  ')

        if self.device is not None:
            b.append(pre[:-1] + 'Device: ' + self.device.name)
            self.device._format(b, pre, name=False)

    def has_extension(self, extension):
        return extension in self.extensions


class OpenCLDeviceInfo:
    """
    Represents information about an OpenCL device.

    An ``OpenCLDeviceInfo`` object has the following properties:

    ``name`` (string)
        This device's name.
    ``vendor`` (string)
        This device's vendor.
    ``version`` (string)
        The OpenCL version supported by this device.
    ``driver`` (string)
        The driver version for this device.
    ``clock`` (int)
        This device's clock speed (in MHz).
    ``globl`` (int)
        The available global memory on this device (in bytes).
    ``local`` (int)
        The available local memory on this device (in bytes).
    ``const`` (int)
        The available constant memory on this device (in bytes).
    ``units`` (int)
        The number of computing units on this device.
    ``param`` (int)
        The maximum total size (in bytes) of arguments passed to the
        kernel. This limits the number of arguments a kernel can get.
    ``groups`` (int)
        The maximum work group size.
    ``dimensions`` (int)
        The maximum work item dimension.
    ``items`` (tuple)
        A tuple of ints specifying the maximum work item size in each
        dimension.

    ``OpenCLDeviceInfo`` objects are created as part of a :class:`OpenCLInfo`
    objects, as returned by most OpenCL enabled parts of Myokit.
    """
    def __init__(self, device):
        self.name = device['name'].strip()
        self.vendor = device['vendor'].strip()
        self.version = device['version'].strip()
        self.driver = device['driver'].strip()
        self.clock = device['clock']
        self.globl = device['global']
        self.local = device['local']
        self.const = device['const']
        self.param = device['param']
        self.groups = device['groups']
        self.items = tuple(device['items'])

    def _format(self, b, pre='', name=True):
        """
        Formats the information in this object and adds it to the list ``b``.
        """
        if name:
            b.append(pre + 'Name            : ' + self.name)
        b.append(pre + 'Vendor          : ' + self.vendor)
        b.append(pre + 'Version         : ' + self.version)
        b.append(pre + 'Driver          : ' + self.driver)
        b.append(pre + 'Clock speed     : ' + str(self.clock) + ' MHz')
        b.append(pre + 'Global memory   : ' + bytesize(self.globl))
        b.append(pre + 'Local memory    : ' + bytesize(self.local))
        b.append(pre + 'Constant memory : ' + bytesize(self.const))
        b.append(pre + 'Max param size  : ' + str(self.param) + ' bytes')
        b.append(pre + 'Max work groups : ' + str(self.groups))
        b.append(pre + 'Max work items  : [' +
                 ', '.join([str(x) for x in self.items]) + ']')


def r(x):
    """Round x, convert to int if possible, the convert to string."""
    return str(myokit.float.round(round(x, 2)))


def bytesize(size):
    """
    Returns a formatted version of a ``size`` given in bytes.
    """
    # Format a size
    if size >= 1073741824:
        return r(0.1 * int(10 * (float(size) / 1073741824))) + ' GB'
    elif size >= 1048576:
        return r(0.1 * int(10 * (float(size) / 1048576))) + ' MB'
    elif size >= 1024:
        return r(0.1 * int(10 * (float(size) / 1024))) + ' KB'
    else:
        return str(size) + ' B'


def clockspeed(speed):
    """
    Returns a formatted version of a ``speed`` given in MHz.
    """
    # Format a size
    if speed >= 1000:
        return r(0.1 * int(10 * (float(speed) / 1000))) + ' GHz'
    else:
        return r(speed) + ' MHz'


class NoOpenCLError(myokit.MyokitError):
    """
    Raised when OpenCLInfo functions requiring OpenCL are called but no OpenCL
    support can be detected.
    """




#!/usr/bin/env python3
#
# Tests protocol creation and pacing
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import pickle
import unittest

import myokit

from myokit.tests import WarningCollector


class ProtocolTest(unittest.TestCase):
    """
    Tests the Protocol class.
    """

    def test_characteristic_time(self):
        # Test characteristic_time determination.

        # Singular event
        e = myokit.ProtocolEvent(1, 100, 0.5, 0, 0)
        self.assertEqual(e.characteristic_time(), 100.5)
        # Finite event
        e = myokit.ProtocolEvent(1, 100, 0.5, 1000, 3)
        self.assertEqual(e.characteristic_time(), 3100)
        # Indefinite event
        e = myokit.ProtocolEvent(1, 100, 0.5, 1000, 0)
        self.assertEqual(e.characteristic_time(), 1000)
        # Delayed indefinite event
        e = myokit.ProtocolEvent(1, 900, 200, 1000, 0)
        self.assertEqual(e.characteristic_time(), 1900)
        # Test protocols
        # Singular event
        p = myokit.Protocol()
        p.schedule(1, 100, 0.5, 0, 0)
        self.assertEqual(p.characteristic_time(), 100.5)
        # Finite event
        p = myokit.Protocol()
        p.schedule(1, 100, 0.5, 1000, 3)
        self.assertEqual(p.characteristic_time(), 3100)
        # Indefinite event
        p = myokit.Protocol()
        p.schedule(1, 100, 0.5, 1000, 0)
        self.assertEqual(p.characteristic_time(), 1000)
        # Delayed indefinite event
        p = myokit.Protocol()
        p.schedule(1, 800, 250, 1000, 0)
        self.assertEqual(p.characteristic_time(), 1800)
        # Sequence of singular events
        p = myokit.Protocol()
        p.schedule(1, 0, 100)
        p.schedule(1, 100, 200)
        p.schedule(1, 300, 300)
        self.assertEqual(p.characteristic_time(), 600)

    def test_clone(self):
        # Test cloning

        p1 = myokit.Protocol()
        p1.schedule(1, 0, 100)
        p1.schedule(1, 100, 200, 1000, 10)
        p1.schedule(1, 300, 300, 2000)
        p2 = p1.clone()
        self.assertEqual(p1, p2)

    def test_equals(self):
        # Test protocol equality checking

        p1 = myokit.Protocol()
        p2 = myokit.Protocol()
        self.assertEqual(p1, p2)

        p1.schedule(1, 0, 100)
        self.assertNotEqual(p1, p2)
        p2.schedule(1, 0, 100)
        self.assertEqual(p1, p2)

        p1.schedule(1, 100, 200, 1000, 10)
        p1.schedule(1, 300, 300, 2000)
        self.assertNotEqual(p1, p2)
        p2.schedule(1, 100, 200, 1000, 10)
        p2.schedule(1, 300, 300, 2000)
        self.assertEqual(p1, p2)

        self.assertNotEqual(p1, None)
        self.assertNotEqual(p1, 17)
        self.assertNotEqual(p1, p1.code())
        self.assertEqual(p1, p1)

    def test_event_creation(self):
        # Test creating events.

        # Invalid event specifications
        def create(start, duration, period=0, multiplier=0):
            myokit.ProtocolEvent(1, start, duration, period, multiplier)
        create(0, 1)
        create(1, 0)
        create(1, 1)
        create(1, 1, 0, 0)
        create(1, 1, 1, 0)
        create(1, 1, 1, 1)
        self.assertRaises(myokit.ProtocolEventError, create, -1, 0)
        self.assertRaises(myokit.ProtocolEventError, create, 0, -1)
        self.assertRaises(myokit.ProtocolEventError, create, -1, 1)
        self.assertRaises(myokit.ProtocolEventError, create, 1, -1)
        self.assertRaises(myokit.ProtocolEventError, create, 0, 0, -1)
        self.assertRaises(myokit.ProtocolEventError, create, 0, 0, 0, 1)
        self.assertRaises(myokit.ProtocolEventError, create, 0, 0, 1, -1)
        self.assertRaises(myokit.ProtocolEventError, create, 0, 0, 1, 0.5)
        self.assertRaises(myokit.ProtocolEventError, create, 0, 2, 1)
        self.assertRaises(myokit.ProtocolEventError, create, 0, 2, 1, 1)

    def test_in_words(self):
        # Test :meth:`Protocol.in_words()`.

        p = myokit.Protocol()
        self.assertEqual(p.in_words(), 'Empty protocol.')

        p = myokit.Protocol()
        p.schedule(2, 10, 100, 1000, 2)
        self.assertEqual(
            p.in_words(),
            'Stimulus of 2.0 times the normal level applied at t=10.0, '
            'lasting 100.0 and occurring 2 times with a period of 1000.0.')

        p = myokit.Protocol()
        p.schedule(2, 10, 100, 1000, 0)
        self.assertEqual(
            p.in_words(),
            'Stimulus of 2.0 times the normal level applied at t=10.0,'
            ' lasting 100.0 and recurring indefinitely'
            ' with a period of 1000.0.')
        self.assertEqual(p.in_words(), str(p.tail()))

        p = myokit.Protocol()
        p.schedule(2, 10, 100, 1000, 0)
        p.schedule(2, 300, 100, 1000, 0)
        self.assertEqual(
            p.in_words(),
            'Stimulus of 2.0 times the normal level applied at t=10.0,'
            ' lasting 100.0 and recurring indefinitely'
            ' with a period of 1000.0.\n'
            'Stimulus of 2.0 times the normal level applied at t=300.0,'
            ' lasting 100.0 and recurring indefinitely'
            ' with a period of 1000.0.')

    def test_is_infinite(self):
        # Tests :meth:`Protocol.is_infinite()

        p = myokit.Protocol()
        self.assertFalse(p.is_infinite())

        p = myokit.Protocol()
        p.schedule(2, 10, 100, 1000, 2)
        self.assertFalse(p.is_infinite())

        p = myokit.Protocol()
        p.schedule(2, 10, 100, 1000, 2)
        p.schedule(2, 1, 1, 1000, 0)
        self.assertTrue(p.is_infinite())

    def test_is_sequence(self):
        # Tests :meth:`Protocol.is_sequence()

        p = myokit.Protocol()
        self.assertTrue(p.is_sequence())
        self.assertTrue(p.is_sequence_exception())
        self.assertTrue(p.is_unbroken_sequence())
        self.assertTrue(p.is_unbroken_sequence_exception())

        # First event is periodic
        p = myokit.Protocol()
        p.schedule(2, 10, 100, 1000, 2)
        self.assertFalse(p.is_sequence())
        self.assertRaisesRegex(
            Exception, 'contains periodic', p.is_sequence_exception)
        self.assertFalse(p.is_unbroken_sequence())
        self.assertRaisesRegex(
            Exception, 'contains periodic', p.is_unbroken_sequence_exception)

        # Second event is periodic
        p = myokit.Protocol()
        p.schedule(2, 10, 100, 0, 0)
        p.schedule(20, 100, 100, 1000, 1000)
        self.assertFalse(p.is_sequence())
        self.assertRaisesRegex(
            Exception, 'contains periodic', p.is_sequence_exception)
        self.assertFalse(p.is_unbroken_sequence())
        self.assertRaisesRegex(
            Exception, 'contains periodic', p.is_unbroken_sequence_exception)

        # Multiple periodic events
        p = myokit.Protocol()
        p.schedule(2, 10, 100, 1000, 2)
        p.schedule(2, 1, 1, 1000, 0)
        self.assertFalse(p.is_sequence())
        self.assertRaisesRegex(
            Exception, 'contains periodic', p.is_sequence_exception)
        self.assertFalse(p.is_unbroken_sequence())
        self.assertRaisesRegex(
            Exception, 'contains periodic', p.is_unbroken_sequence_exception)

        # Overlapping events
        p = myokit.Protocol()
        p.schedule(2, 10, 100, 0, 0)
        p.schedule(2, 12, 1, 0, 0)
        p.schedule(2, 420, 1, 0, 0)
        self.assertFalse(p.is_sequence())
        self.assertRaisesRegex(
            Exception, 'overlap', p.is_sequence_exception)
        self.assertFalse(p.is_unbroken_sequence())
        self.assertRaisesRegex(
            Exception, 'overlap', p.is_unbroken_sequence_exception)

        # Non-overlapping events
        p = myokit.Protocol()
        p.schedule(2, 10, 100, 0, 0)
        p.schedule(2, 120, 1, 0, 0)
        p.schedule(2, 420, 1, 0, 0)
        self.assertTrue(p.is_sequence())
        self.assertTrue(p.is_sequence_exception())
        self.assertFalse(p.is_unbroken_sequence())
        self.assertRaisesRegex(
            Exception, 'start directly afte', p.is_unbroken_sequence_exception)

        # Unbroken sequence
        p = myokit.Protocol()
        p.schedule(2, 10, 100, 0, 0)
        p.schedule(2, 110, 310, 0, 0)
        p.schedule(2, 420, 1, 0, 0)
        self.assertTrue(p.is_sequence())
        self.assertTrue(p.is_sequence_exception())
        self.assertTrue(p.is_unbroken_sequence())
        self.assertTrue(p.is_unbroken_sequence_exception())

    def test_levels(self):
        # Tests :meth:`Protocol.levels()

        p = myokit.Protocol()
        p.schedule(2, 10, 100, 0, 0)
        p.schedule(3, 110, 310, 0, 0)
        p.schedule(1, 420, 1, 0, 0)
        self.assertTrue(p.is_unbroken_sequence())
        self.assertEqual(p.levels(), [2, 3, 1])

    def test_log_for_interval(self):
        # Tests the method Protocol.log_for_interval()
        # Relies on PacingSystem

        debug = False

        # Test basic use + log creation methods
        p = myokit.Protocol()
        p.schedule(1, 10, 1, 1000, 0)
        d = p.log_for_interval(0, 3000)
        self.assertEqual(d.time(), [0, 10, 11, 1010, 1011, 2010, 2011, 3000])
        d = p.log_for_interval(0, 2000, for_drawing=True)
        if debug:
            import matplotlib.pyplot as plt
            plt.figure()
            plt.plot(d.time(), d['pace'])
            plt.show()
        self.assertEqual(d.time(), [
            0, 10, 10, 11, 11, 1010, 1010, 1011, 1011, 2000])
        p = myokit.Protocol()
        p.schedule(1, 0, 1, 1000, 0)
        d = p.log_for_interval(0, 3000)
        self.assertEqual(d.time(), [0, 1, 1000, 1001, 2000, 2001, 3000])
        d = p.log_for_interval(0, 2000, for_drawing=True)
        if debug:
            import matplotlib.pyplot as plt
            plt.figure()
            plt.plot(d.time(), d['pace'])
            plt.show()
        self.assertEqual(
            d.time(), [0, 1, 1, 1000, 1000, 1001, 1001, 2000, 2000])

        # Test bad interval call
        self.assertRaises(ValueError, p.log_for_interval, 100, 0)

        # Test deprecated alias
        with WarningCollector() as w:
            p.create_log_for_interval(0, 2000, for_drawing=True)
        self.assertIn('deprecated', w.text())

    def test_log_for_times(self):
        # Test the method Protocol.log_for_times()
        # Relies on PacingSystem

        p = myokit.Protocol()
        p.schedule(2, 10, 100, 1000, 2)

        t = [0, 9.999, 10, 10.001, 109.999, 110, 110.001, 1000, 1009.99, 1010,
             1110, 2000, 2020]
        v = [0, 0, 2, 2, 2, 0, 0, 0, 0, 2, 0, 0, 0]
        d = p.log_for_times(t)
        self.assertEqual(len(d), 2)
        self.assertIn('time', d)
        self.assertIn('pace', d)
        self.assertEqual(d.time(), t)
        self.assertEqual(d['pace'], v)

        # Empty times
        d = p.log_for_times([])
        self.assertEqual(len(d.time()), 0)
        self.assertEqual(len(d['pace']), 0)

        # Deprecated alias
        with WarningCollector() as w:
            p.create_log_for_times([])
        self.assertIn('deprecated', w.text())

    def test_pickling(self):
        # Test protocol pickling

        p1 = myokit.Protocol()
        p1.schedule(1, 0, 100)
        p1.schedule(1, 100, 200, 1000, 10)
        p1.schedule(1, 300, 300, 2000)

        p_bytes = pickle.dumps(p1)
        p2 = pickle.loads(p_bytes)
        self.assertEqual(p1, p2)

    def test_protocol_creation(self):
        # Test the basics of creating a protocol
        p = myokit.Protocol()
        self.assertIsNone(p.head())
        p.schedule(1, 10, 0.5, 0, 0)
        f = p.head()
        self.assertEqual(f.level(), 1)
        self.assertEqual(f.start(), 10)
        self.assertEqual(f.duration(), 0.5)
        self.assertEqual(f.period(), 0)
        self.assertEqual(f.multiplier(), 0)
        self.assertEqual(f.stop(), 10.5)

        # Invalid event: period is zero but multiplier is not
        self.assertRaises(
            myokit.ProtocolEventError, p.schedule, 1, 10, 0.5, 0, 10)

        # Add second event
        p.schedule(2, 1, 0.5, 0, 0)
        e = p.head()
        self.assertNotEqual(e, f)
        self.assertEqual(e.level(), 2)
        self.assertEqual(e.start(), 1)
        self.assertEqual(e.duration(), 0.5)
        self.assertEqual(e.period(), 0)
        self.assertEqual(e.multiplier(), 0)

        # Add third event
        p.schedule(3, 100, 0.5, 100, 100)

        # Invalid event: starts simultaneously
        def sim(p, start, duration, period=0, multiplier=0, clash=0):
            with self.assertRaises(myokit.SimultaneousProtocolEventError) as e:
                p.schedule(2, start, duration, period, multiplier)
            m = str(e.exception)
            t = m[2 + m.index('t='):-1]
            self.assertEqual(float(t), clash)
        sim(p, 1, 0.5, clash=1)
        sim(p, 10, 0.5, clash=10)
        sim(p, 100, 0.5, clash=100)

    def test_protocol_to_string(self):
        # Tests str(protocol)
        p = myokit.Protocol()
        p.schedule(2, 10, 100, 0, 0)
        p.schedule(3, 110, 310, 0, 0)
        p.schedule(1, 420, 1, 0, 0)
        self.assertEqual(p.code(), str(p))

    def test_range(self):
        # Tests :meth:`Protocol.range()
        p = myokit.Protocol()
        self.assertEqual(p.range(), (0, 0))

        p = myokit.Protocol()
        p.schedule(2, 10, 100, 0, 0)
        p.schedule(3, 110, 310, 0, 0)
        p.schedule(1, 420, 1, 0, 0)
        self.assertTrue(p.is_unbroken_sequence())
        self.assertEqual(p.range(), (0, 3))

        p = myokit.Protocol()
        p.schedule(2, 0, 110, 0, 0)
        p.schedule(3, 110, 310, 0, 0)
        p.schedule(1, 420, 1, 0, 0)
        self.assertTrue(p.is_unbroken_sequence())
        self.assertEqual(p.range(), (1, 3))

    def test_tail(self):
        # Tests Protocol.tail() which returns the final protocol event

        p = myokit.Protocol()
        p.schedule(2, 10, 100, 0, 0)
        p.schedule(3, 110, 310, 0, 0)
        p.schedule(1, 420, 1, 0, 0)
        self.assertEqual(p.tail().start(), 420)

        p = myokit.Protocol()
        p.schedule(1, 420, 1, 0, 0)
        p.schedule(2, 10, 100, 0, 0)
        p.schedule(3, 110, 310, 0, 0)
        self.assertEqual(p.tail().start(), 420)

    def test_value_at_times(self):
        # Test the method Protocol.value_at_times()
        # Relies on PacingSystem

        p = myokit.Protocol()
        p.schedule(2, 10, 100, 1000, 2)
        t = [0, 9.999, 10, 10.001, 109.999, 110, 110.001, 1000, 1009.99, 1010,
             1110, 2000, 2020]
        v = [0, 0, 2, 2, 2, 0, 0, 0, 0, 2, 0, 0, 0]
        self.assertEqual(v, p.value_at_times(t))

        # Empty times
        d = p.log_for_times([])
        self.assertEqual(len(p.value_at_times([])), 0)

        # Decreasing times
        self.assertRaisesRegex(
            ValueError, 'non-decreasing', p.value_at_times, [1, 0])

        # Negative times
        self.assertRaisesRegex(
            ValueError, 'negative', p.value_at_times, [-1, 0])


if __name__ == '__main__':
    import warnings
    warnings.simplefilter('always')
    unittest.main()



#!/usr/bin/env python3
#
# Tests the CModel class.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import myokit

from myokit.tests import DIR_DATA


class CModelTest(unittest.TestCase):
    """
    Tests the CModel class.
    """

    @classmethod
    def setUpClass(cls):
        # Test cmodel instantiation.
        m, p, x = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        cls.model = m
        cls.sensitivities = (['ik1.gK1', 'ikp.IKp'], ['cell.K_o', 'ikp.gKp'])
        cls.pacing_labels = {}
        cls.cmodel = myokit.CModel(
            cls.model, cls.pacing_labels, cls.sensitivities
        )

    def test_sensitivities(self):
        # Test instantiation of cmodel with sensitivities

        # Bad type
        sens = 'Bad type'
        with self.assertRaisesRegex(ValueError, 'The argument `sensitivities'):
            myokit.CModel(self.model, self.pacing_labels, sens)

        # Empty deps or indeps
        sens = ([], ['some parameter'])
        m = myokit.CModel(self.model, self.pacing_labels, sens)
        self.assertFalse(m.has_sensitivities)
        sens = (['some state'], [])
        m = myokit.CModel(self.model, self.pacing_labels, sens)
        self.assertFalse(m.has_sensitivities)

        # Provide sensitivies as Variables
        s1 = self.model.get('ik1.gK1')
        s2 = self.model.get('ikp.IKp')
        p1 = self.model.get('cell.K_o')
        p2 = self.model.get('ikp.gKp')
        sens = ([s1, s2], [p1, p2])
        m = myokit.CModel(self.model, self.pacing_labels, sens)
        self.assertTrue(m.has_sensitivities)

        # Provide sensitivities as Names
        sens = (
            [myokit.Name(s1), myokit.Name(s2)],
            [myokit.Name(p1), myokit.Name(p2)])
        m = myokit.CModel(self.model, self.pacing_labels, sens)
        self.assertTrue(m.has_sensitivities)

        # Sensitivity of derivative
        s3 = self.model.get('ik.x')
        sens = (
            [myokit.Derivative(myokit.Name(s3)), myokit.Name(s2)],
            [myokit.Name(p1), myokit.Name(p2)])
        m = myokit.CModel(self.model, self.pacing_labels, sens)
        self.assertTrue(m.has_sensitivities)
        s3 = 'dot(ik.x)'
        sens = (
            [s3, myokit.Name(s2)],
            [myokit.Name(p1), myokit.Name(p2)])
        m = myokit.CModel(self.model, self.pacing_labels, sens)
        self.assertTrue(m.has_sensitivities)

        # Sensitivity of derivative of non-state
        sens = (
            [myokit.Derivative(myokit.Name(s1)), myokit.Name(s2)],
            [myokit.Name(p1), myokit.Name(p2)])
        with self.assertRaisesRegex(ValueError, 'Sensitivity of '):
            myokit.CModel(self.model, self.pacing_labels, sens)

        # Sensitivity of bound variable
        sens = (
            ['engine.time', myokit.Name(s2)],
            [myokit.Name(p1), myokit.Name(p2)])
        with self.assertRaisesRegex(ValueError, 'Sensitivities cannot'):
            myokit.CModel(self.model, self.pacing_labels, sens)

        # Sensitivity w.r.t. Initial value
        s3 = self.model.get('ik.x')
        sens = (
            [s3, myokit.Name(s2)],
            [myokit.Name(p1), myokit.InitialValue(myokit.Name(s3))])
        m = myokit.CModel(self.model, self.pacing_labels, sens)
        self.assertTrue(m.has_sensitivities)

        # Sensitivity w.r.t. initial value of non-state
        sens = (
            [myokit.Name(s1), myokit.Name(s2)],
            [myokit.Name(p1), myokit.InitialValue(myokit.Name(p2))])
        with self.assertRaisesRegex(ValueError, 'Sensitivity with respect to'):
            myokit.CModel(self.model, self.pacing_labels, sens)

        # Sensitivity w.r.t. non-literal
        sens = (
            [myokit.Name(s1), myokit.Name(s2)],
            [myokit.Name(p1), 'ik.E'])
        with self.assertRaisesRegex(ValueError, 'Sensitivity with respect to'):
            myokit.CModel(self.model, self.pacing_labels, sens)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Class for testing the event-based pacing system.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os

import myokit

from myokit.tests import DIR_TEST


# Location of C template
SOURCE_FILE = 'ansic_event_based_pacing.c'


class AnsicEventBasedPacing(myokit.CModule):
    """
    Class for testing the event-based pacing system.
    """
    _index = 0

    def __init__(self, protocol, initial_time=0):
        super().__init__()

        # Unique id
        AnsicEventBasedPacing._index += 1
        module_name = 'myokit_ansic_pacing_' \
            + str(AnsicEventBasedPacing._index)

        # Arguments
        fname = os.path.join(DIR_TEST, SOURCE_FILE)

        # Create back-end
        args = {'module_name': module_name}
        libs = []
        libd = []
        incd = [DIR_TEST, myokit.DIR_CFUNC]
        self._sys = None
        self._sys = self._compile(module_name, fname, args, libs, libd, incd)

        # Initialize
        self._sys.init(protocol.clone(), initial_time)
        self.advance(initial_time)

    def __del__(self):
        # Free the memory used by the pacing system
        if self._sys is not None:
            self._sys.clean()

    def advance(self, new_time):
        """
        Advances the time in the pacing system to ``new_time``.

        Returns the current value of the pacing variable.
        """
        return self._sys.advance(new_time)

    def next_time(self):
        """
        Returns the next time the value of the pacing variable will be
        updated.
        """
        return self._sys.next_time()

    def pace(self):
        """
        Returns the current value of the pacing variable.
        """
        return self._sys.pace()

    def time(self):
        """
        Returns the current time in the pacing system.
        """
        return self._sys.time()

    @staticmethod
    def log_for_interval(protocol, a, b):
        """
        Copied from PacingSystem. This is useful for testing!
        """
        # Test the input
        a, b = float(a), float(b)
        if b < a:
            raise ValueError('The argument `b` cannot be smaller than `a`')

        # Create a simulation log
        log = myokit.DataLog()
        log.set_time_key('time')
        log['time'] = time = []
        log['pace'] = pace = []

        # Create a pacing system
        p = AnsicEventBasedPacing(protocol)

        # Fill in points
        t = a
        v = p.advance(t)
        time.append(t)
        pace.append(v)
        while t < b:
            t = min(p.next_time(), b)
            v = p.advance(t)
            time.append(t)
            pace.append(v)
        return log




#!/usr/bin/env python3
#
# Tests the CellML 1.0/1.1 parser.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import myokit
import myokit.formats.cellml as cellml
import myokit.formats.cellml.v1 as v1

from myokit.tests import TemporaryDirectory, DIR_FORMATS, WarningCollector

# CellML directory
DIR = os.path.join(DIR_FORMATS, 'cellml')


class TestCellMLParser(unittest.TestCase):
    """ Tests the CellML 1.0/1.1 parser (mostly for errors). """

    def assertBad(self, xml, message, version='1.0'):
        """
        Inserts the given ``xml`` into a <model> element, parses it, and checks
        that this raises an exception matching ``message``.
        """
        self.assertRaisesRegex(
            v1.CellMLParsingError, message, self.parse, xml, version)

    def parse(self, xml, version='1.0'):
        """
        Inserts the given ``xml`` into a <model> element, parses it, and
        returns the result.
        """
        return v1.parse_string(self.wrap(xml, version))

    def parse_in_file(self, xml):
        """
        Inserts the given ``xml`` into a <model> element, writes it to a
        temporary file, parses it, and returns the result.
        """
        with TemporaryDirectory() as d:
            path = d.path('test.cellml')
            with open(path, 'w') as f:
                f.write(self.wrap(xml))
            return v1.parse_file(path)

    def wrap(self, xml, version='1.0'):
        """
        Wraps the given ``xml`` into a <model> element.
        """
        assert version in ('1.0', '1.1'), 'version must be 1.0 or 1.1'
        v = version
        # Note: Meta data stays version 1.0

        return (
            '<?xml version="1.0" encoding="UTF-8"?>'
            '<model name="test"'
            '       xmlns="http://www.cellml.org/cellml/' + v + '#"'
            '       xmlns:cellml="http://www.cellml.org/cellml/' + v + '#"'
            '       xmlns:cmeta="http://www.cellml.org/metadata/1.0#"'
            '       xmlns:rdf="' + cellml.NS_RDF + '"'
            '       xmlns:bqbiol="' + cellml.NS_BQBIOL + '">'
            + xml +
            '</model>')

    def test_cellml_namespace(self):
        # Attributes from the CellML namespace are not allowed

        self.assertBad(
            '<component name="c" cellml:name="d" />',
            'Unexpected attribute cellml:name',
        )

    def test_cmeta_ids(self):
        # Test cmeta ids are parsed

        # Test parsing cmeta id
        path = os.path.join(DIR, 'br-1977.cellml')
        model = v1.parse_file(path)
        self.assertEqual(model.cmeta_id(), 'beeler_reuter_1977')

        # Invalid cmeta id
        self.assertBad(
            '<component cmeta:id="" name="c" />',
            'non-empty string')

        # Duplicate cmeta id
        self.assertBad(
            '<component cmeta:id="x" name="c" />'
            '<component cmeta:id="x" name="d" />',
            'Duplicate cmeta:id')

    def test_cmeta_namespace(self):
        # Only the cmeta:id attribute is allowed, no other cmeta attributes or
        # elements.

        # Attribute in cmeta
        self.assertBad(
            '<component name="c" cmeta:ernie="bert" />',
            'Unexpected attribute cmeta:ernie')

        # Element in cmeta
        self.assertBad(
            '<component name="c">'
            '  <cmeta:bert />'
            '</component>',
            'element of type cmeta:bert')

    def test_component(self):
        # Test component parsing

        # Parse component
        m = self.parse('<component name="ernie" />')
        self.assertIn('ernie', m)

        # Component must have name
        self.assertBad('<component />', 'Component element must have a name')

        # CellML errors are converted
        self.assertBad('<component name="1" />', 'identifier')

        # Reactions are not supported
        self.assertBad(
            '<component name="c">'
            '<reaction />'
            '</component>',
            'Reactions are not supported')

        # Component can have units
        m = self.parse(
            '<component name="ernie">'
            '<units name="wooster">'
            '<unit units="meter" />'
            '</units>'
            '</component>'
        )
        e = m['ernie']
        u = e.find_units('wooster')

    def test_connection(self):
        # Test connection parsing

        x = ('<component name="a">'
             '  <variable name="x" units="volt" public_interface="in" />'
             '</component>'
             '<component name="b">'
             '  <variable name="y" units="volt" public_interface="out"'
             '            initial_value="1" />'
             '</component>')

        # Parse valid connection
        y = ('<connection>'
             '  <map_components component_1="a" component_2="b" />'
             '  <map_variables variable_1="x" variable_2="y" />'
             '</connection>')
        m = self.parse(x + y)
        self.assertIs(m['a']['x'].source(), m['b']['y'])
        self.assertIsNone(m['b']['y'].source())

        # No map components
        self.assertBad(x + '<connection />', 'exactly one map_components')

        # No map variables
        y = ('<connection>'
             '  <map_components component_1="a" component_2="b" />'
             '</connection>')
        self.assertBad(x + y, 'at least one map_variables')

    def test_connection_map_components(self):
        # Test parsing map_components elements

        x = ('<component name="a">'
             '  <variable name="x" units="volt" public_interface="in" />'
             '</component>'
             '<component name="b">'
             '  <variable name="y" units="volt" public_interface="out" />'
             '</component>')

        # Valid connection is already checked

        # Missing component 1 or 2
        y = ('<connection>'
             '  <map_components component_2="b" />'
             '  <map_variables variable_1="x" variable_2="y" />'
             '</connection>')
        self.assertBad(x + y, 'must define a component_1 attribute')
        y = ('<connection>'
             '  <map_components component_1="a" />'
             '  <map_variables variable_1="x" variable_2="y" />'
             '</connection>')
        self.assertBad(x + y, 'must define a component_2 attribute')

        # Identical components
        y = ('<connection>'
             '  <map_components component_1="a" component_2="a" />'
             '  <map_variables variable_1="x" variable_2="y" />'
             '</connection>')
        self.assertBad(x + y, 'must be different')

        # Non-existent components
        y = ('<connection>'
             '  <map_components component_1="ax" component_2="b" />'
             '  <map_variables variable_1="x" variable_2="y" />'
             '</connection>')
        self.assertBad(x + y, 'component_1 attribute must refer to')
        y = ('<connection>'
             '  <map_components component_1="a" component_2="bx" />'
             '  <map_variables variable_1="x" variable_2="y" />'
             '</connection>')
        self.assertBad(x + y, 'component_2 attribute must refer to')

        # Connected twice
        y = ('<connection>'
             '  <map_components component_1="a" component_2="b" />'
             '  <map_variables variable_1="x" variable_2="y" />'
             '</connection>')
        self.assertBad(x + y + y, 'unique pair of components')

    def test_connection_map_variables(self):
        # Test parsing map_variables elements

        x = ('<component name="a">'
             '  <variable name="x" units="volt" public_interface="in" />'
             '</component>'
             '<component name="b">'
             '  <variable name="y" units="volt" public_interface="out" />'
             '</component>'
             '<connection>'
             '  <map_components component_1="a" component_2="b" />')
        z = ('</connection>')

        # Valid connection is already checked

        # Missing variable 1 or 2
        y = '<map_variables variable_2="y" />'
        self.assertBad(x + y + z, 'must define a variable_1 attribute')
        y = '<map_variables variable_1="x" />'
        self.assertBad(x + y + z, 'must define a variable_2 attribute')

        # Non-existent variable 1 & 2
        y = '<map_variables variable_1="y" variable_2="y" />'
        self.assertBad(x + y + z, 'variable_1 attribute must refer to')
        y = '<map_variables variable_1="x" variable_2="z" />'
        self.assertBad(x + y + z, 'variable_2 attribute must refer to')

        # Connecting twice is not OK
        y = '<map_variables variable_1="x" variable_2="y" />'
        self.assertRaisesRegex(
            v1.CellMLParsingError, 'already connected to',
            self.parse, x + y + y + z)

        # Bad interfaces etc. propagate from cellml API
        q = ('<component name="a">'
             '  <variable name="x" units="volt" public_interface="out" />'
             '</component>'
             '<component name="b">'
             '  <variable name="y" units="volt" public_interface="out" />'
             '</component>'
             '<connection>'
             '  <map_components component_1="a" component_2="b" />'
             '  <map_variables variable_1="x" variable_2="y" />'
             '</connection>')
        self.assertRaisesRegex(
            v1.CellMLParsingError, 'Invalid connection',
            self.parse, q)

    def test_documentation(self):
        # Tests parsing a <documentation> tag from the cellml temp doc ns

        path = os.path.join(DIR, 'documentation.cellml')
        expected = '\n'.join([
            'Article title',
            '',
            'Michael',
            'Clerx',
            '',
            'University of Oxford',
            '',
            'Model Status',
            'This model is fake.',
            '',
            'This is a paragraph.',
            '',
            'schematic diagram',
            '',
            'A schematic diagram describing the model.',
            '',
            'Here\'s some extra documentation.',
        ])
        m = v1.parse_file(path)
        self.assertEqual(m.meta['documentation'], expected)

    def test_evaluated_derivatives(self):
        # Test parsing a simple model; compare RHS derivatives to known ones

        # Load myokit model
        org_model = myokit.load_model('example')
        org_states = [x.qname() for x in org_model.states()]
        org_values = org_model.evaluate_derivatives()

        # Load exported version
        path = os.path.join(DIR, 'lr-1991-exported-1.cellml')
        cm = v1.parse_file(path)
        new_model = cm.myokit_model()
        new_states = [x.qname() for x in new_model.states()]
        new_values = new_model.evaluate_derivatives()

        # Compare each state (loop unrolled for easier debugging)
        org_i = 0
        new_i = new_states.index(org_states[org_i])
        self.assertEqual(org_values[org_i], new_values[new_i])

        org_i = 1
        new_i = new_states.index(org_states[org_i])
        self.assertEqual(org_values[org_i], new_values[new_i])

        org_i = 2
        new_i = new_states.index(org_states[org_i])
        self.assertEqual(org_values[org_i], new_values[new_i])

        org_i = 3
        new_i = new_states.index(org_states[org_i])
        self.assertEqual(org_values[org_i], new_values[new_i])

        org_i = 4
        new_i = new_states.index(org_states[org_i])
        self.assertEqual(org_values[org_i], new_values[new_i])

        org_i = 5
        new_i = new_states.index(org_states[org_i])
        self.assertEqual(org_values[org_i], new_values[new_i])

        org_i = 6
        new_i = new_states.index(org_states[org_i])
        self.assertEqual(org_values[org_i], new_values[new_i])

        org_i = 7
        new_i = new_states.index(org_states[org_i])
        self.assertEqual(org_values[org_i], new_values[new_i])

    def test_extensions(self):
        # Test handling of extension elements

        # CellML elements can not appear inside extensions
        self.assertBad(
            '<x:y xmlns:x="x"><component name="c" /></x:y>',
            'found inside extension element')
        self.assertBad(
            '<x:a xmlns:x="xxx">'
            '  <y:b xmlns:y="yyy">'
            '    <z:b xmlns:z="zzz">'
            '      <component name="c" />'
            '    </z:b>'
            '  </y:b>'
            '</x:a>',
            'found inside extension element')

        # CellML attributes can not appear inside extensions
        self.assertBad(
            '<x:y xmlns:x="x" cellml:name="a" />',
            'found in extension element')
        self.assertBad(
            '<x:a xmlns:x="xxx">'
            '  <y:b xmlns:y="yyy">'
            '    <z:b xmlns:z="zzz" cellml:name="b" />'
            '  </y:b>'
            '</x:a>',
            'found in extension element')

    def test_group(self):
        # Tests parsing a group element.

        x = ('<component name="a" />'
             '<component name="b" />'
             '<component name="c" />'
             )

        # Parse valid group
        y = ('<group>'
             '  <relationship_ref relationship="encapsulation"/>'
             '  <component_ref component="a">'
             '    <component_ref component="b">'
             '      <component_ref component="c" />'
             '    </component_ref>'
             '  </component_ref>'
             '</group>')
        m = self.parse(x + y)
        self.assertEqual(m['c'].parent(), m['b'])
        self.assertEqual(m['b'].parent(), m['a'])
        self.assertIsNone(m['a'].parent())

        # Missing component ref
        y = ('<group>'
             '  <relationship_ref relationship="encapsulation"/>'
             '</group>')
        self.assertBad(x + y, 'at least one component_ref')

        # Missing relationship ref
        y = ('<group>'
             '  <component_ref component="a">'
             '    <component_ref component="b" />'
             '  </component_ref>'
             '</group>')
        self.assertBad(x + y, 'at least one relationship_ref')

        # In encapsulation and containment relationships, the first
        # component_ref must have at least one child
        y = ('<group>'
             '  <relationship_ref relationship="encapsulation" />'
             '  <component_ref component="a" />'
             '</group>')
        self.assertBad(x + y, r'must have at least one child \(6.4.3.2\)')
        y = ('<group>'
             '  <relationship_ref relationship="containment" />'
             '  <component_ref component="a" />'
             '</group>')
        self.assertBad(x + y, r'must have at least one child \(6.4.3.2\)')

        # But it's fine for extension types
        y = ('<group xmlns:x="x">'
             '  <relationship_ref x:relationship="family" />'
             '  <component_ref component="a" />'
             '</group>')
        self.parse(x + y)

    def test_group_component_ref(self):
        # Tests parsing a component_ref element.

        # Valid has already been checked

        x = ('<component name="a" />'
             '<component name="b" />'
             '<component name="c" />'
             '<group>'
             '  <relationship_ref relationship="encapsulation"/>'
             '  <component_ref component="a">'
             '    <component_ref component="b" />')
        z = ('  </component_ref>'
             '</group>')

        # Missing component attribute
        y = '<component_ref />'
        self.assertBad(x + y + z, 'must define a component attribute')

        # Non-existent component
        y = '<component_ref component="jane" />'
        self.assertBad(x + y + z, 'must reference a component')

        # Two parents
        y = ('</component_ref>'
             '<component_ref component="c">'
             '  <component_ref component="b" />')
        self.assertBad(x + y + z, 'only have a single encapsulation parent')

        # CellML errors propagate
        y = ('</component_ref>'
             '<component_ref component="b">'
             '  <component_ref component="a" />')
        self.assertBad(x + y + z, 'hierarchy cannot be circular')

    def test_group_relationship_ref(self):
        # Tests parsing a relationsip_ref element.

        # Valid has already been checked

        x = ('<component name="a" />'
             '<component name="b" />'
             '<group>'
             '  <component_ref component="a">'
             '    <component_ref component="b" />'
             '  </component_ref>')
        z = ('</group>')

        # Containment is ok but ignored
        y = '<relationship_ref relationship="containment" />'
        m = self.parse(x + y + z)
        self.assertIsNone(m['a'].parent())
        self.assertIsNone(m['b'].parent())

        # Relationship in other namespace is allowed but ignored
        y = ('<relationship_ref'
             '   xmlns:zippy="http://example.com"'
             '   zippy:relationship="a-loose-affiliation" />')
        m = self.parse(x + y + z)
        self.assertIsNone(m['a'].parent())
        self.assertIsNone(m['b'].parent())

        # Only containment and encapsulation exist within CellML
        y = '<relationship_ref relationship="equal" />'
        self.assertBad(x + y + z, 'Unknown relationship type')

        # Relationship attribute must be present
        y = '<relationship_ref />'
        self.assertBad(x + y + z, 'must define a relationship attribute')

        # Multiple relationship types are allowed
        y = ('<relationship_ref relationship="encapsulation" />'
             '<relationship_ref relationship="containment" />')
        m = self.parse(x + y + z)
        self.assertIsNone(m['a'].parent())
        self.assertEqual(m['b'].parent(), m['a'])

        # Multiple named containment relationship types are allowed
        y = ('<relationship_ref relationship="containment" name="x" />'
             '<relationship_ref relationship="containment" />')
        m = self.parse(x + y + z)
        self.assertIsNone(m['a'].parent())
        self.assertIsNone(m['b'].parent())

        # Duplicate relationships are not allowed
        y = ('<relationship_ref relationship="containment" />'
             '<relationship_ref relationship="containment" />')
        self.assertBad(x + y + z, 'must have a unique pair')
        y = ('<relationship_ref relationship="containment" name="xx" />'
             '<relationship_ref relationship="containment" name="xx" />')
        self.assertBad(x + y + z, 'must have a unique pair')

        # Name must be a valid identifier
        y = '<relationship_ref relationship="containment" name="123" />'
        self.assertBad(x + y + z, 'valid CellML identifier')

        # Encapsulation can't be named
        y = '<relationship_ref relationship="encapsulation" name="betty" />'
        self.assertBad(x + y + z, 'may not define a name')

    def test_import(self):
        # Import elements are not supported

        # Not allowed in 1.0
        self.assertBad(
            '<import />', 'not allowed in CellML 1.0', version='1.0')

        # Not supported in 1.1
        self.assertBad(
            '<import />', 'Imports are not supported', version='1.1')

    def test_math(self):
        # Tests parsing math elements
        x = ('<component name="a">'
             '  <variable name="x" units="volt" />'
             '  <math xmlns="http://www.w3.org/1998/Math/MathML">')
        z = ('  </math>'
             '</component>'
             '<component name="b">'
             '  <variable name="y" units="volt" initial_value="12.3" />'
             '</component>')

        # Test parsing valid equation
        y = '<apply><eq /><ci>x</ci><cn cellml:units="volt">-80</cn></apply>'
        m = self.parse(x + y + z)
        var = m['a']['x']
        self.assertEqual(var.rhs(), myokit.Number(-80, myokit.units.volt))

        # Valid equation inside a semantics element
        m = self.parse(x + '<semantics>' + y + '</semantics>' + z)
        var = m['a']['x']
        self.assertEqual(var.rhs(), myokit.Number(-80, myokit.units.volt))

        # Valid equation with some annotations to ignore
        m = self.parse(x + '<annotation />' + y + '<annotation-xml />' + z)
        var = m['a']['x']
        self.assertEqual(var.rhs(), myokit.Number(-80, myokit.units.volt))

        # Constants
        m = self.parse(x + '<apply><eq /><ci>x</ci> <pi /> </apply>' + z)
        var = m['a']['x']
        self.assertEqual(var.rhs().unit(), myokit.units.dimensionless)

        # Variable doesn't exist
        y = '<apply><eq /><ci>y</ci><cn cellml:units="volt">-80</cn></apply>'
        self.assertBad(x + y + z, 'Variable references in equation must name')
        y = '<apply><eq /><ci>x</ci><ci>y</ci></apply>'
        self.assertBad(x + y + z, 'Variable references in equation must name')

        # No units in number
        y = '<apply><eq /><ci>x</ci><cn>-80</cn></apply>'
        self.assertBad(x + y + z, 'must define a cellml:units')

        # Non-existent units
        y = '<apply><eq /><ci>x</ci><cn cellml:units="vlop">-80</cn></apply>'
        self.assertBad(x + y + z, 'Unknown unit "vlop" referenced')

        # Unsupported units: Warning
        y = '<apply><eq /><ci>x</ci><cn cellml:units="celsius">2</cn></apply>'
        with WarningCollector() as w:
            self.parse(x + y + z)
        self.assertIn('celsius', w.text())

        # Items outside of MathML namespace
        y = '<cellml:component name="bertie" />'
        self.assertBad(x + y + z, 'must be in the mathml namespace')

        # Doesn't start with an apply
        y = '<ci>x</ci>'
        self.assertBad(x + y + z, 'Expecting mathml:apply but found')

        # Not an equality
        y = '<apply><plus /><ci>x</ci></apply>'
        self.assertBad(x + y + z, 'expecting a list of equations')

        # Not in assignment form
        y = ('<apply>'
             ' <eq />'
             ' <apply><plus /><ci>x</ci><cn cellml:units="volt">1</cn></apply>'
             ' <cn cellml:units="volt">10</cn>'
             '</apply>')
        self.assertBad(x + y + z, 'Invalid expression found on the left-hand')

        # Assignment to a variable with an "in" interface (error should be
        # converted from CellMLError to CellMLParsingError).
        x = ('<component name="a">'
             '  <variable name="x" units="volt" public_interface="in" />'
             '  <math xmlns="http://www.w3.org/1998/Math/MathML">')
        y = '<apply><eq /><ci>x</ci><cn cellml:units="volt">-80</cn></apply>'
        self.assertBad(x + y + z, 'public_interface="in"')

        # MathML inside the model element
        self.assertBad(
            '<math xmlns="http://www.w3.org/1998/Math/MathML" />',
            'element of type mathml:math',
        )

        # Attributes from the MathML namespace
        self.assertBad(
            '<component name="c" mathml:name="d" '
            ' xmlns:mathml="http://www.w3.org/1998/Math/MathML" />',
            'Unexpected attribute mathml:name',
        )

    def test_maths_1_1(self):
        # Test setting a variable as initial value, allowed in 1.1

        # Legal case in 1.1
        x = ('<component name="a">'
             '  <variable name="p" units="volt" initial_value="q" />'
             '  <variable name="q" units="volt" initial_value="12.3" />'
             '</component>')

        # Not allowed in 1.0
        self.assertBad(x, r'a real number \(3.4.3.7\)', version='1.0')

        # Not supported in 1.1
        self.assertBad(x, 'not supported', version='1.1')

    def test_model(self):
        # Tests parsing a model element.

        # Valid one is already tested

        # Invalid namespace
        x = '<?xml version="1.0" encoding="UTF-8"?>'
        y = '<model name="x" xmlns="http://example.com" />'
        self.assertRaisesRegex(
            v1.CellMLParsingError, 'must be in CellML',
            v1.parse_string, x + y)

        # CellML 1.0 and 1.1 are ok
        y = '<model name="x" xmlns="http://www.cellml.org/cellml/1.0#" />'
        m = v1.parse_string(x + y)
        self.assertEqual(m.version(), '1.0')
        y = '<model name="x" xmlns="http://www.cellml.org/cellml/1.1#" />'
        m = v1.parse_string(x + y)
        self.assertEqual(m.version(), '1.1')

        # Not a model
        y = '<module name="x" xmlns="http://www.cellml.org/cellml/1.0#" />'
        self.assertRaisesRegex(
            v1.CellMLParsingError, 'must be a CellML model',
            v1.parse_string, x + y)

        # No name
        y = '<model xmlns="http://www.cellml.org/cellml/1.0#" />'
        self.assertRaisesRegex(
            v1.CellMLParsingError, 'Model element must have a name',
            v1.parse_string, x + y)

        # CellML API errors are wrapped
        y = '<model name="123" xmlns="http://www.cellml.org/cellml/1.0#" />'
        self.assertRaisesRegex(
            v1.CellMLParsingError, 'valid CellML identifier',
            v1.parse_string, x + y)

        # Too many free variables
        self.assertBad(
            '<component name="a">'
            '  <variable name="x" units="dimensionless" initial_value="0.1" />'
            '  <variable name="y" units="dimensionless" initial_value="2.3" />'
            '  <variable name="time1" units="dimensionless" />'
            '  <variable name="time2" units="dimensionless" />'
            '  <math xmlns="http://www.w3.org/1998/Math/MathML">'
            '    <apply>'
            '      <eq />'
            '      <apply>'
            '        <diff/>'
            '        <bvar>'
            '          <ci>time1</ci>'
            '        </bvar>'
            '        <ci>x</ci>'
            '      </apply>'
            '      <cn cellml:units="dimensionless">1</cn>'
            '    </apply>'
            '    <apply>'
            '      <eq />'
            '      <apply>'
            '        <diff/>'
            '        <bvar>'
            '          <ci>time2</ci>'
            '        </bvar>'
            '        <ci>y</ci>'
            '      </apply>'
            '      <cn cellml:units="dimensionless">2</cn>'
            '    </apply>'
            '  </math>'
            '</component>',
            'derivatives with respect to more than one variable')

    def test_overdefined(self):
        # Tests for overdefined variables

        # Initial value and equation for state
        x = ('<component name="a">'
             '  <variable name="time" units="dimensionless" />'
             '  <variable name="x" units="dimensionless" initial_value="2" />'
             '  <math xmlns="http://www.w3.org/1998/Math/MathML">')
        a = ('    <apply>'
             '      <eq />'
             '      <apply>'
             '        <diff/>'
             '        <bvar>'
             '          <ci>time</ci>'
             '        </bvar>'
             '        <ci>x</ci>'
             '      </apply>'
             '      <cn cellml:units="dimensionless">3</cn>'
             '    </apply>')
        z = ('  </math>'
             '</component>')
        m = self.parse(x + a + z)
        v = m['a']['x']
        self.assertEqual(v.rhs(), myokit.Number(3))
        self.assertEqual(v.initial_value(), 2)

        # Initial value and equation for a non-state
        b = ('    <apply>'
             '      <eq /><ci>x</ci><cn cellml:units="dimensionless">2</cn>'
             '    </apply>')
        self.assertBad(x + b + z, 'Initial value and a defining equation')

        # Two equations
        self.assertBad(x + a + a + z, 'Two defining equations')

    def test_oxmeta_annotations(self):
        # Tests parsing of RDF annotations for the Web Lab

        # Start and end of required tags
        r1 = '<rdf:RDF>'
        r2 = '</rdf:RDF>'
        d1a = '<rdf:Description rdf:about="#'
        d1b = '">'
        d2 = '</rdf:Description>'
        b1 = '<bqbiol:is rdf:resource="' + cellml.NS_OXMETA
        b2 = '"/>'

        # Create mini model with 2 rdf tags and 3 annotations
        x = ('<component name="a">'
             '  <variable name="x" units="volt" cmeta:id="x"'
             '            initial_value="1" />'
             '  <variable name="y" units="volt" cmeta:id="yzie"'
             '            initial_value="1" />'
             '  <variable name="z" units="volt" cmeta:id="zed"'
             '            initial_value="1" />'
             '</component>')
        y = r1
        y += d1a + 'x' + d1b + b1 + 'membrane_voltage' + b2 + d2
        y += d1a + 'yzie' + d1b + b1 + 'sodium_reversal_potential' + b2 + d2
        y += r2
        y += r1
        y += d1a + 'zed' + d1b + b1 + 'calcium_reversal_potential' + b2 + d2
        y += r2

        # Parse
        cm = self.parse(x + y)

        # Check cmeta ids
        x, y, z = cm['a']['x'], cm['a']['y'], cm['a']['z']
        self.assertEqual(x.cmeta_id(), 'x')
        self.assertEqual(y.cmeta_id(), 'yzie')
        self.assertEqual(z.cmeta_id(), 'zed')

        # Check oxmeta annotation
        self.assertIn('oxmeta', x.meta)
        self.assertIn('oxmeta', y.meta)
        self.assertIn('oxmeta', z.meta)
        self.assertEqual(x.meta['oxmeta'], 'membrane_voltage')
        self.assertEqual(y.meta['oxmeta'], 'sodium_reversal_potential')
        self.assertEqual(z.meta['oxmeta'], 'calcium_reversal_potential')

    def test_oxmeta_annotations_bad(self):
        # Tests parsing of RDF annotations for the Web Lab

        # Start and end of required tags
        r1 = '<rdf:RDF>'
        r2 = '</rdf:RDF>'
        d1a = '<rdf:Description rdf:about="#'
        d1b = '">'
        d2 = '</rdf:Description>'
        b1 = '<bqbiol:is rdf:resource="' + cellml.NS_OXMETA
        b2 = '"/>'

        # Model code
        x = ('<component name="a">'
             '  <variable name="x" units="volt" cmeta:id="x"'
             '            initial_value="1" />'
             '</component>')

        # Check that parser survives all kinds of half-formed annotations

        # No content
        y = r1 + r2
        cm = self.parse(x + y)

        # No about attribute
        y = r1 + '<rdf:Description />' + r2
        cm = self.parse(x + y)

        # Unknown cmeta id
        y = r1 + d1a + 'xxx' + d1b + d2 + r2
        cm = self.parse(x + y)

        # No bqbiol:is
        y = r1 + d1a + 'x' + d1b + d2 + r2
        cm = self.parse(x + y)

        # No bqbiol:is
        y = r1 + d1a + 'x' + d1b + d2 + r2
        cm = self.parse(x + y)

        # No resource
        y = r1 + d1a + 'x' + d1b + '<bqbiol:is/>' + d2 + r2
        cm = self.parse(x + y)

        # Non oxmeta-resource
        y = r1 + d1a + 'x' + d1b + '<bqbiol:is rdf:resource="hi"/>' + d2 + r2
        cm = self.parse(x + y)

    def test_parse_file(self):
        # Tests the parse file method

        # Parse a valid file
        m = self.parse_in_file('<component name="a" />')
        self.assertIn('a', m)

        # Parse invalid XML: errors must be wrapped
        self.assertRaisesRegex(
            v1.CellMLParsingError,
            'Unable to parse XML',
            self.parse_in_file,
            '<component',
        )

    def test_parse_string(self):
        # Tests the parse string method

        # Valid version is already tested

        # Parse invalid XML: errors must be wrapped
        self.assertRaisesRegex(
            v1.CellMLParsingError,
            'Unable to parse XML',
            v1.parse_string,
            'Hello there',
        )

    def test_rdf(self):
        # rdf:RDF is allowed, but no other elements or attributes

        # rdf:RDF is ok
        self.parse('<rdf:RDF />')

        # Attribute from rdf namespace
        self.assertBad(
            '<component name="c" rdf:RDF="d" />',
            'Unexpected attribute rdf:RDF')

        # Element from rdf namespace
        self.assertBad(
            '<rdf:robert />',
            'found element of type rdf:robert')

    def test_text_in_elements(self):
        # Test for text inside (and after) elements

        self.assertBad(
            '<component name="c">Hiya</component>',
            'Text found in cellml:component')
        self.assertBad(
            '<component name="c"></component>Hiya',
            'Text found in cellml:model')

    def test_unexpected(self):
        # Test for unexpected elements and attributes

        self.assertBad('<wooster />', 'Unexpected content type')
        self.assertBad(
            '<component name="c" food="nice" />', 'Unexpected attribute')

    def test_unit(self):
        # Test parsing a unit row

        # Parse a valid set of units (that needs sorting)
        x = ('<units name="megawooster">'
             '  <unit units="wooster" prefix="mega" />'
             '</units>'
             '<units name="wooster">'
             '  <unit units="volt"'
             '        prefix="milli"'
             '        exponent="2"'
             '        multiplier="1.2" />'
             '</units>')
        m = self.parse(x)
        wooster_ref = m.find_units('wooster')
        wooster_myo = ((myokit.units.volt * 1e-3) ** 2) * 1.2
        self.assertEqual(wooster_myo, wooster_ref.myokit_unit())
        megawooster_ref = m.find_units('megawooster')
        self.assertEqual(wooster_myo * 1e6, megawooster_ref.myokit_unit())

        # Zero offsets are OK
        x = ('<units name="wooster">'
             '  <unit units="volt" offset="0.0" />'
             '</units>')
        m = self.parse(x)
        self.assertEqual(
            m.find_units('wooster').myokit_unit(), myokit.units.volt)

        # Non-zero offsets are not supported: treated as dimensionless
        x = ('<units name="nonzero">'
             '  <unit units="meter" />'
             '  <unit units="volt" offset="0.1" />'
             '</units>')
        with WarningCollector() as w:
            m = self.parse(x)
        self.assertIn('non-zero offsets are not supported', w.text())
        self.assertEqual(
            myokit.units.dimensionless, m.find_units('nonzero').myokit_unit())

        # Non-integer exponents are supported
        x = ('<units name="unsup">'
             '  <unit units="ampere" exponent="2.34" />'
             '</units>')
        m = self.parse(x)
        self.assertEqual(
            m.find_units('unsup').myokit_unit(),
            myokit.units.ampere**2.34,
        )

        # Offset must be a number
        x = ('<units name="wooster">'
             '  <unit units="volt" offset="hello" />'
             '</units>')
        self.assertBad(x, 'must be a real number')

        # Test that CellML errors are wrapped
        x = ('<units name="wooster">'
             '  <unit units="volt" prefix="woopsie" />'
             '</units>')
        self.assertBad(x, 'must be a string from the list')

        # No units definition
        x = ('<units name="wooster">'
             '  <unit />'
             '</units>')
        self.assertBad(x, 'must have a units attribute')

        # Unsupported base units raise a warning, are treated as dimensionless
        x = ('<units name="wooster">'
             '  <unit units="meter" />'
             '  <unit units="celsius" />'
             '</units>')
        with WarningCollector() as w:
            m = self.parse(x)
        self.assertIn('celsius', w.text())
        self.assertEqual(
            myokit.units.dimensionless, m.find_units('wooster').myokit_unit())

    def test_units(self):
        # Test parsing a units definition

        # Valid has already been testsed

        # Base units attribute is allowed be no
        x = ('<units name="wooster" base_units="no">'
             '  <unit units="volt" />'
             '</units>')
        m = self.parse(x)
        self.assertEqual(
            m.find_units('wooster').myokit_unit(), myokit.units.volt)

        # Base units attribute must be yes or no
        x = ('<units name="wooster" base_units="bjork">'
             '  <unit units="volt" offset="0.0" />'
             '</units>')
        self.assertBad(x, 'must be either "yes" or "no"')

        # New base units are not supported
        x = '<units name="base" base_units="yes" />'
        with WarningCollector() as w:
            m = self.parse(x)
        self.assertIn('new base unit', w.text())
        self.assertEqual(
            m.find_units('base').myokit_unit(), myokit.units.dimensionless)

        # CellML errors are converted to parse errors
        x = '<units name="123"><unit units="volt" /></units>'
        self.assertBad(x, 'valid CellML identifier')

        # Missing name
        x = '<units><unit units="volt" /></units>'
        self.assertBad(x, 'must have a name')

        # Name overlaps with predefined
        x = '<units name="meter"><unit units="volt" /></units>'
        self.assertBad(x, 'overlaps with a predefined name')

        # Same error for celsius (unsupported units)
        x = '<units name="meter"><unit units="celsius" /></units>'
        self.assertBad(x, 'overlaps with a predefined name')

        # Duplicate name (handled in sorting)
        x = '<units name="wooster"><unit units="volt" /></units>'
        self.assertBad(x + x, 'Duplicate units definition')

        # No child unit elements
        x = '<units name="woopster" />'
        self.assertBad(x, 'at least one child unit element')

        # Missing units definitions
        x = ('<units name="wooster"><unit units="fluther" /></units>')
        self.assertBad(x, 'Unable to resolve network of units')

        # Circular units definitions
        x = ('<units name="wooster"><unit units="fluther" /></units>'
             '<units name="fluther"><unit units="wooster" /></units>')
        self.assertBad(x, 'Unable to resolve network of units')

        # Test unit resolving for shadowed units
        x = ('<units name="wooster"><unit units="ampere" /></units>'
             '<component name="a">'
             '  <units name="kilowooster">'
             '    <unit units="wooster" prefix="kilo" />'
             '  </units>'
             '  <units name="wooster"><unit units="volt" /></units>'
             '</component>')
        m = self.parse(x)
        u = m['a'].find_units('kilowooster').myokit_unit()
        self.assertEqual(u, myokit.units.volt * 1000)

    def test_units_predefined(self):
        # Tests parsing all the predefined units

        def u(units):
            m = self.parse(
                '<component name="c">'
                '  <variable name="x" units="' + units + '"'
                '   initial_value="1" />'
                '</component>'
            )
            return m['c']['x'].units().myokit_unit()

        self.assertEqual(u('ampere'), myokit.units.ampere)
        self.assertEqual(u('becquerel'), myokit.units.becquerel)
        self.assertEqual(u('candela'), myokit.units.candela)
        self.assertEqual(u('coulomb'), myokit.units.coulomb)
        self.assertEqual(u('dimensionless'), myokit.units.dimensionless)
        self.assertEqual(u('farad'), myokit.units.farad)
        self.assertEqual(u('gram'), myokit.units.g)
        self.assertEqual(u('gray'), myokit.units.gray)
        self.assertEqual(u('henry'), myokit.units.henry)
        self.assertEqual(u('hertz'), myokit.units.hertz)
        self.assertEqual(u('joule'), myokit.units.joule)
        self.assertEqual(u('katal'), myokit.units.katal)
        self.assertEqual(u('kelvin'), myokit.units.kelvin)
        self.assertEqual(u('kilogram'), myokit.units.kg)
        self.assertEqual(u('liter'), myokit.units.liter)
        self.assertEqual(u('litre'), myokit.units.liter)
        self.assertEqual(u('lumen'), myokit.units.lumen)
        self.assertEqual(u('lux'), myokit.units.lux)
        self.assertEqual(u('meter'), myokit.units.meter)
        self.assertEqual(u('metre'), myokit.units.meter)
        self.assertEqual(u('mole'), myokit.units.mole)
        self.assertEqual(u('newton'), myokit.units.newton)
        self.assertEqual(u('ohm'), myokit.units.ohm)
        self.assertEqual(u('pascal'), myokit.units.pascal)
        self.assertEqual(u('radian'), myokit.units.radian)
        self.assertEqual(u('second'), myokit.units.second)
        self.assertEqual(u('siemens'), myokit.units.siemens)
        self.assertEqual(u('sievert'), myokit.units.sievert)
        self.assertEqual(u('steradian'), myokit.units.steradian)
        self.assertEqual(u('tesla'), myokit.units.tesla)
        self.assertEqual(u('volt'), myokit.units.volt)
        self.assertEqual(u('watt'), myokit.units.watt)
        self.assertEqual(u('weber'), myokit.units.weber)

        # Celsius is not supported
        with WarningCollector() as w:
            self.assertEqual(u('celsius'), myokit.units.dimensionless)
        self.assertIn('celsius', w.text())

    def test_variable(self):
        # Tests parsing variables

        # Valid has already been tested

        # Must have a name
        x = '<component name="a"><variable units="volt"/></component>'
        self.assertBad(x, 'must have a name attribute')

        # Must have units
        x = '<component name="a"><variable name="a" /></component>'
        self.assertBad(x, 'must have a units attribute')

        # CellML errors are converted to parsing errors
        x = '<component name="a"><variable name="1" units="volt"/></component>'
        self.assertBad(x, 'valid CellML identifier')

        # Unsupported units are ignored
        x = '<variable name="x" units="celsius" initial_value="3" />'
        x = '<component name="a">' + x + '</component>'
        with WarningCollector() as w:
            x = self.parse(x)['a']['x']
        self.assertIn('celsius', w.text())

        self.assertEqual(x.units().myokit_unit(), myokit.units.dimensionless)


if __name__ == '__main__':
    import warnings
    warnings.simplefilter('always')
    unittest.main()



#!/usr/bin/env python3
#
# Tests the Fiber-tissue OpenCL simulation
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import numpy as np

import myokit

from myokit.tests import (
    CancellingReporter,
    DIR_DATA,
    OpenCL_FOUND,
    WarningCollector,
)

# Show simulation output
debug = False


@unittest.skipIf(not OpenCL_FOUND, 'OpenCL not found on this system.')
class FiberTissueSimulationTest(unittest.TestCase):
    """
    Tests the fiber-tissue simulation.
    """

    @classmethod
    def setUpClass(cls):
        # Create objects for shared used in testing

        # Fiber model, tissue model, protocol
        cls.mf = myokit.load_model(
            os.path.join(DIR_DATA, 'dn-1985-normalised.mmt'))
        cls.mt = myokit.load_model(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        cls.p = myokit.pacing.blocktrain(1000, 2.0, offset=.01)

        # Fiber/Tissue sizes (small sim)
        cls.nfx = 8
        cls.nfy = 4
        cls.ntx = 8
        cls.nty = 6

        # Shared simulations
        cls._s0 = cls._s1 = None

    @property
    def s0(self):
        # Shared 2x1 simulation
        if self._s0 is None:
            self._s0 = myokit.FiberTissueSimulation(
                self.mt,
                self.mt,
                self.p,
                ncells_fiber=(1, 1),
                ncells_tissue=(1, 1),
                nx_paced=1,
                g_fiber=(0, 0),
                g_tissue=(0, 0),
                g_fiber_tissue=0,
                precision=myokit.DOUBLE_PRECISION,
                dt=0.01,
            )
        return self._s0

    @property
    def s1(self):
        # Shared small simulation
        if self._s1 is None:
            self._s1 = myokit.FiberTissueSimulation(
                self.mf,
                self.mt,
                self.p,
                ncells_fiber=(self.nfx, self.nfy),
                ncells_tissue=(self.ntx, self.nty),
                nx_paced=4,
                g_fiber=(235, 100),
                g_tissue=(9, 5),
                g_fiber_tissue=9,
                dt=0.0012,
            )
        return self._s1

    def test_against_cvode(self):
        # Compare the fiber-tissue simulation output with CVODE output

        # Set up logging
        logvars = ['engine.time', 'engine.pace', 'membrane.V', 'ica.ICa']
        logt = myokit.LOG_NONE

        # Run simulation
        tmax = 100
        dlog = 0.1
        try:
            d1, logt = self.s0.run(
                tmax, logf=logvars, logt=logt, log_interval=dlog)
        finally:
            self.s0.reset()
        del logt
        d1 = d1.npview()

        # Run CVODE simulation
        s2 = myokit.Simulation(self.mt, self.p)
        s2.set_tolerance(1e-8, 1e-8)
        d2 = s2.run(tmax, logvars, log_interval=dlog).npview()

        # Check implementation of logging point selection
        e0 = np.max(np.abs(d1.time() - d2.time()))

        # Check implementation of pacing
        r1 = d1['engine.pace'] - d2['engine.pace']
        e1 = np.sum(r1**2)

        # Check membrane potential (will have some error!)
        # Using MRMS from Marsh, Ziaratgahi, Spiteri 2012
        r2 = d1['membrane.V', 0, 0] - d2['membrane.V']
        r2 /= (1 + np.abs(d2['membrane.V']))
        e2 = np.sqrt(np.sum(r2**2) / len(r2))

        # Check logging of intermediary variables
        r3 = d1['ica.ICa', 0, 0] - d2['ica.ICa']
        r3 /= (1 + np.abs(d2['ica.ICa']))
        e3 = np.sqrt(np.sum(r3**2) / len(r3))

        if debug:
            import matplotlib.pyplot as plt
            print('Event at t=0')

            print(d1.time()[:7])
            print(d2.time()[:7])
            print(d1.time()[-7:])
            print(d2.time()[-7:])
            print(e0)

            plt.figure()
            plt.suptitle('Pacing signals')
            plt.subplot(2, 1, 1)
            plt.plot(d1.time(), d1['engine.pace'], '--', label='FiberTissue')
            plt.plot(d2.time(), d2['engine.pace'], '--', label='CVODE')
            plt.legend()
            plt.subplot(2, 1, 2)
            plt.plot(d1.time(), r1)
            print(e1)

            plt.figure()
            plt.suptitle('Membrane potential')
            plt.subplot(2, 1, 1)
            plt.plot(d1.time(), d1['membrane.V', 0, 0], label='FiberTissue')
            plt.plot(d2.time(), d2['membrane.V'], '--', label='CVODE')
            plt.legend()
            plt.subplot(2, 1, 2)
            plt.plot(d1.time(), r2)
            print(e2)

            plt.figure()
            plt.suptitle('Calcium current')
            plt.subplot(2, 1, 1)
            plt.plot(d1.time(), d1['ica.ICa', 0, 0], label='FiberTissue')
            plt.plot(d2.time(), d2['ica.ICa'], '--', label='CVODE')
            plt.legend()
            plt.subplot(2, 1, 2)
            plt.plot(d1.time(), r2)
            print(e3)

            plt.show()

        self.assertLess(e0, 1e-10)
        self.assertLess(e1, 1e-14)
        self.assertLess(e2, 0.05)
        self.assertLess(e3, 0.01)

    def test_basic(self):
        # Test basic functionality
        s = self.s1
        nfx, nfy, ntx, nty = self.nfx, self.nfy, self.ntx, self.nty

        # Run times
        run = .1

        # Set up logging
        logf = [
            'engine.time',
            'membrane.V',
            'isi.isiCa',
        ]
        logt = [
            'membrane.V',
            'ica.Ca_i',
            'ica.ICa',
        ]
        # Run simulation
        try:
            logf, logt = s.run(run, logf=logf, logt=logt, log_interval=0.01)
        finally:
            s.reset()

        self.assertEqual(len(logf), 1 + 2 * nfx * nfy)
        self.assertIn('engine.time', logf)
        self.assertIn('0.0.membrane.V', logf)
        self.assertIn(str(nfx - 1) + '.' + str(nfy - 1) + '.membrane.V', logf)
        self.assertIn('0.0.isi.isiCa', logf)
        self.assertIn(str(nfx - 1) + '.' + str(nfy - 1) + '.isi.isiCa', logf)

        self.assertEqual(len(logt), 3 * ntx * nty)
        self.assertIn('0.0.membrane.V', logt)
        self.assertIn(str(ntx - 1) + '.' + str(nty - 1) + '.membrane.V', logt)
        self.assertIn('0.0.ica.Ca_i', logt)
        self.assertIn(str(ntx - 1) + '.' + str(nty - 1) + '.ica.Ca_i', logt)
        self.assertIn('0.0.ica.ICa', logt)
        self.assertIn(str(ntx - 1) + '.' + str(nty - 1) + '.ica.ICa', logt)

    def test_creation(self):
        # Tests fiber tisue simulation creation
        mf, mt, p = self.mf, self.mt, self.p
        FT = myokit.FiberTissueSimulation

        # Models must be valid
        m2 = mf.clone()
        m2.label('membrane_potential').set_rhs(None)
        self.assertFalse(m2.is_valid())
        self.assertRaises(myokit.MissingRhsError, FT, m2, mt)
        self.assertFalse(m2.is_valid())
        self.assertRaises(myokit.MissingRhsError, FT, mf, m2)

        # Models must have interdependent components
        m2 = mf.clone()
        x = m2.get('ik').add_variable('xx')
        x.set_rhs('membrane.i_ion')
        self.assertTrue(m2.has_interdependent_components())
        self.assertRaisesRegex(ValueError, 'interdependent', FT, m2, mt)
        self.assertTrue(m2.has_interdependent_components())
        self.assertRaisesRegex(ValueError, 'interdependent', FT, mf, m2)

        # Dimensionalities must be 2d tuples
        s = FT(mf, mt, p, (5, 6), (7, 8))
        self.assertEqual(s.fiber_shape(), (6, 5))
        self.assertEqual(s.tissue_shape(), (8, 7))
        self.assertRaisesRegex(
            ValueError, r'fiber size must be a tuple \(nx, ny\)',
            FT, mf, mt, ncells_fiber=None)
        self.assertRaisesRegex(
            ValueError, r'fiber size must be a tuple \(nx, ny\)',
            FT, mf, mt, ncells_fiber=12)
        self.assertRaisesRegex(
            ValueError, r'fiber size must be a tuple \(nx, ny\)',
            FT, mf, mt, ncells_fiber=(12, ))
        self.assertRaisesRegex(
            ValueError, r'fiber size must be a tuple \(nx, ny\)',
            FT, mf, mt, ncells_fiber=(12, 12, 12))

        self.assertRaisesRegex(
            ValueError, r'tissue size must be a tuple \(nx, ny\)',
            FT, mf, mt, ncells_tissue=None)
        self.assertRaisesRegex(
            ValueError, r'tissue size must be a tuple \(nx, ny\)',
            FT, mf, mt, ncells_tissue=4)
        self.assertRaisesRegex(
            ValueError, r'tissue size must be a tuple \(nx, ny\)',
            FT, mf, mt, ncells_tissue=(3, ))
        self.assertRaisesRegex(
            ValueError, r'tissue size must be a tuple \(nx, ny\)',
            FT, mf, mt, ncells_tissue=(6, 6, 6))

        # Number of cells must be at least 1
        self.assertRaisesRegex(
            ValueError, r'fiber size must be at least \(1, 1\)',
            FT, mf, mt, ncells_fiber=(0, 10))
        self.assertRaisesRegex(
            ValueError, r'fiber size must be at least \(1, 1\)',
            FT, mf, mt, ncells_fiber=(10, 0))
        self.assertRaisesRegex(
            ValueError, r'fiber size must be at least \(1, 1\)',
            FT, mf, mt, ncells_fiber=(-1, -1))
        self.assertRaisesRegex(
            ValueError, r'tissue size must be at least \(1, 1\)',
            FT, mf, mt, ncells_tissue=(0, 10))
        self.assertRaisesRegex(
            ValueError, r'tissue size must be at least \(1, 1\)',
            FT, mf, mt, ncells_tissue=(10, 0))
        self.assertRaisesRegex(
            ValueError, r'tissue size must be at least \(1, 1\)',
            FT, mf, mt, ncells_tissue=(-1, -1))
        self.assertRaisesRegex(
            ValueError, 'exceed that of the tissue',
            FT, mf, mt, p, (5, 6), (5, 5))

        # Number of paced cells must be >= 0
        FT(mf, mt, p, nx_paced=0)
        self.assertRaisesRegex(
            ValueError, 'width of the stimulus pulse must be non-negative',
            FT, mf, mt, p, nx_paced=-1)

        # Check conductivities are tuples
        self.assertRaisesRegex(
            ValueError, r'fiber conductivity must be a tuple \(gx, gy\)',
            FT, mf, mt, p, g_fiber=1)
        self.assertRaisesRegex(
            ValueError, r'fiber conductivity must be a tuple \(gx, gy\)',
            FT, mf, mt, p, g_fiber=(1, ))
        self.assertRaisesRegex(
            ValueError, r'fiber conductivity must be a tuple \(gx, gy\)',
            FT, mf, mt, p, g_fiber=(1, 1, 1))
        self.assertRaisesRegex(
            ValueError, r'tissue conductivity must be a tuple \(gx, gy\)',
            FT, mf, mt, p, g_tissue=1)
        self.assertRaisesRegex(
            ValueError, r'tissue conductivity must be a tuple \(gx, gy\)',
            FT, mf, mt, p, g_tissue=(1, ))
        self.assertRaisesRegex(
            ValueError, r'tissue conductivity must be a tuple \(gx, gy\)',
            FT, mf, mt, p, g_tissue=(1, 1, 1))

        # Check step size is > 0
        self.assertRaisesRegex(
            ValueError, 'step size must be greater', FT, mf, mt, p, dt=0)
        self.assertRaisesRegex(
            ValueError, 'step size must be greater', FT, mf, mt, p, dt=-1)

        # Precision must be single or double
        self.assertRaisesRegex(
            ValueError, 'Only single and double',
            FT, mf, mt,
            precision=myokit.SINGLE_PRECISION + myokit.DOUBLE_PRECISION)

        # Membrane potentials must be given with labels
        m2 = mf.clone()
        m2.label('membrane_potential').set_label(None)
        self.assertRaisesRegex(
            ValueError, '"membrane_potential" in the fiber model',
            FT, m2, mt)
        self.assertRaisesRegex(
            ValueError, '"membrane_potential" in the tissue model',
            FT, mf, m2)

        # Membrane potential must be a state
        m2.get('ik.iK').set_label('membrane_potential')
        self.assertRaisesRegex(
            ValueError, 'fiber model must be a state variable',
            FT, m2, mt)
        self.assertRaisesRegex(
            ValueError, 'tissue model must be a state variable',
            FT, mf, m2)

        # Membrane potential vars must have same unit
        m2 = mf.clone()
        m2.label('membrane_potential').set_unit(None)
        self.assertRaisesRegex(
            ValueError, 'fiber model must specify a unit for the membrane',
            FT, m2, mt)
        self.assertRaisesRegex(
            ValueError, 'tissue model must specify a unit for the membrane',
            FT, mf, m2)
        m2.label('membrane_potential').set_unit(myokit.units.pF)
        self.assertRaisesRegex(
            ValueError, 'membrane potential must have the same unit',
            FT, mf, m2)

        # Diffusion currents must be given with labels
        m2 = mf.clone()
        m2.binding('diffusion_current').set_binding(None)
        self.assertRaisesRegex(
            ValueError, 'fiber model to be bound to "diffusion_current"',
            FT, m2, mt)
        self.assertRaisesRegex(
            ValueError, 'tissue model to be bound to "diffusion_current"',
            FT, mf, m2)

        # Diffusion vars must have same unit
        m2 = mf.clone()
        m2.binding('diffusion_current').set_unit(None)
        self.assertRaisesRegex(
            ValueError, 'fiber model must specify a unit for the diffusion',
            FT, m2, mt)
        self.assertRaisesRegex(
            ValueError, 'tissue model must specify a unit for the diffusion',
            FT, mf, m2)
        m2.binding('diffusion_current').set_unit(myokit.units.pF)
        self.assertRaisesRegex(
            ValueError, 'diffusion current must have the same unit',
            FT, mf, m2)

    def test_negative_time(self):
        # Test starting at a negative time

        p = myokit.pacing.blocktrain(duration=1, level=1, period=2)
        self.s0.reset()
        self.s0.set_protocol(p)
        self.s0.set_time(-10)
        self.assertEqual(self.s0.time(), -10)
        df, dt = self.s0.run(5, logf=['engine.pace'], logt=myokit.LOG_NONE)
        df = df.npview()
        self.assertEqual(self.s0.time(), -5)
        self.assertTrue(np.all(df['engine.pace'] == 0))
        df, dt = self.s0.run(
            6, logf=['engine.time', 'engine.pace'], logt=myokit.LOG_NONE)
        df = df.npview()
        #print(d['engine.time'])
        #print(d['engine.pace'])
        self.assertEqual(self.s0.time(), 1)
        self.assertTrue(np.all(df['engine.pace'][:-1] == 0))
        self.assertEqual(df['engine.pace'][-1], 1)

    def test_pre_reset(self):
        # Tests getting/setting of states and time, and use of pre() and run()
        try:
            # Check setting of time
            self.assertEqual(self.s1.time(), 0)
            self.s1.set_time(10)
            self.assertEqual(self.s1.time(), 10)
            self.s1.reset()     # Check that reset() resets time
            self.assertEqual(self.s1.time(), 0)

            # Check initial state equals default state
            sf, st = self.s1.fiber_state(), self.s1.tissue_state()
            self.assertEqual(sf, self.s1.default_fiber_state())
            self.assertEqual(st, self.s1.default_tissue_state())
        finally:
            self.s1.reset()

        # At this point, we have either quit or sf and st are set.
        # Use these to restore the original sim state at the end of the test.
        sf0, st0 = sf, st

        try:
            # Check running updates time and states, but not default states
            self.s1.run(5, logf=myokit.LOG_NONE, logt=myokit.LOG_NONE)
            self.assertEqual(self.s1.time(), 5)
            self.assertNotEqual(
                self.s1.fiber_state(), self.s1.default_fiber_state())
            self.assertNotEqual(
                self.s1.tissue_state(), self.s1.default_tissue_state())
            self.assertEqual(sf, self.s1.default_fiber_state())
            self.assertEqual(st, self.s1.default_tissue_state())

            # Setting time to 0 is different from reset()
            self.s1.set_time(0)
            self.assertEqual(self.s1.time(), 0)
            self.assertNotEqual(
                self.s1.fiber_state(), self.s1.default_fiber_state())
            self.assertNotEqual(
                self.s1.tissue_state(), self.s1.default_tissue_state())

            # Check reset() fixes everything
            self.s1.reset()
            self.assertEqual(sf, self.s1.fiber_state())
            self.assertEqual(st, self.s1.tissue_state())
            self.assertEqual(self.s1.time(), 0)

            # Check that pre() updates both states, but not time
            self.s1.pre(5)
            self.assertEqual(self.s1.time(), 0)
            self.assertNotEqual(sf, self.s1.fiber_state())
            self.assertNotEqual(st, self.s1.tissue_state())
            sf, st = self.s1.fiber_state(), self.s1.tissue_state()
            self.assertEqual(sf, self.s1.default_fiber_state())
            self.assertEqual(st, self.s1.default_tissue_state())
            self.assertEqual(self.s1.time(), 0)

            # Check that reset restores new default state
            self.s1.run(5, logf=myokit.LOG_NONE, logt=myokit.LOG_NONE)
            self.assertEqual(self.s1.time(), 5)
            self.assertNotEqual(
                self.s1.fiber_state(), self.s1.default_fiber_state())
            self.assertNotEqual(
                self.s1.tissue_state(), self.s1.default_tissue_state())
            self.assertEqual(sf, self.s1.default_fiber_state())
            self.assertEqual(st, self.s1.default_tissue_state())
            self.s1.reset()
            self.assertEqual(self.s1.time(), 0)
            self.assertEqual(sf, self.s1.fiber_state())
            self.assertEqual(st, self.s1.tissue_state())

        finally:
            self.s1.set_default_fiber_state(sf0)
            self.s1.set_default_tissue_state(st0)
            self.s1.reset()

    def test_protocol(self):
        # Test setting a protocol

        try:
            # Run, check Vm is raised
            df, dt = self.s0.run(
                5, logf=['membrane.V'], logt=myokit.LOG_NONE, log_interval=1)
            self.assertGreater(df['membrane.V', 0, 0][-1], 0)
            self.s0.reset()

            # Run with constant proto
            x = 0.0123
            self.s0.set_protocol(
                myokit.pacing.blocktrain(level=x, duration=1000, period=1000))
            self.s0.reset()
            df, dt = self.s0.run(3, logf=['engine.pace'])
            df = df.npview()
            self.assertTrue(np.all(df['engine.pace'] == x))

            # Unset: Must now be zero
            self.s0.set_protocol(None)
            self.s0.reset()
            df, dt = self.s0.run(3, logf=['engine.pace'])
            df = df.npview()
            self.assertTrue(np.all(df['engine.pace'] == 0))

            # Reset protocol, check Vm goes high
            self.s0.set_protocol(self.p)
            self.s0.reset()
            df, dt = self.s0.run(
                5, logf=['membrane.V'], logt=myokit.LOG_NONE, log_interval=1)
            self.assertGreater(df['membrane.V', 0, 0][-1], 0)
        finally:
            self.s0.reset()

    def test_run_errors(self):
        # Test invalid calls to run()

        try:
            # Run with negative sim time
            self.assertRaisesRegex(ValueError, 'can\'t be negative',
                                   self.s1.run, -1)

            # Run with negative or zero log interval: Just gets set to really
            # small value
            self.s1.run(1, log_interval=0)
            self.s1.run(1, log_interval=-1)
        finally:
            self.s1.reset()

    def test_run_progress(self):
        # Tests running with a progress reporter.
        pass

        # Run with a progress reporter
        with myokit.tools.capture() as c:
            self.s1.run(3, progress=myokit.ProgressPrinter())
        self.assertTrue(c.text() != '')

        # Cancel using a progress reporter
        self.assertRaises(
            myokit.SimulationCancelledError,
            self.s1.run, 20, progress=CancellingReporter(0))

        # Run with invalid progress reporter
        try:
            self.assertRaisesRegex(ValueError, 'subclass of myokit.ProgressR',
                                   self.s1.run, 1, progress=12)
        finally:
            self.s1.reset()

    def test_fiber_state(self):
        # Test the fiber_state related methods.

        # Check simulation state equals model state
        m = self.mf.count_states()
        nx, ny = self.nfx, self.nfy

        sm = self.mf.initial_values(True)
        for i in range(nx):
            for j in range(ny):
                self.assertEqual(sm, self.s1.fiber_state(i, j))

        try:
            # Test setting a full-sized state
            sx = list(range(nx * ny * m))
            self.s1.set_fiber_state(sx)
            self.assertEqual(sx, self.s1.fiber_state())

            # Test setting a single, global state
            sx = [0.0] * m
            self.assertNotEqual(sm, sx)
            self.s1.set_fiber_state(sx)
            for i in range(nx):
                for j in range(ny):
                    self.assertEqual(sx, self.s1.fiber_state(i, j))
            self.assertEqual(sx * (nx * ny), self.s1.fiber_state())
            self.s1.set_fiber_state(sm)
            self.assertEqual(sm * (nx * ny), self.s1.fiber_state())

            # Test setting the state of a single cell
            x, y = 1, 2
            self.s1.set_fiber_state(sx, x, y)
            for i in range(nx):
                for j in range(ny):
                    if i == x and j == y:
                        self.assertEqual(self.s1.fiber_state(i, j), sx)
                    else:
                        self.assertEqual(self.s1.fiber_state(i, j), sm)

            # Test indexing in state vector is x first, then y
            self.s1.set_fiber_state(sx)
            self.s1.set_fiber_state(sm, x=1, y=2)
            s = self.s1.fiber_state()[::m]
            self.assertEqual(s[2 * nx + 1], sm[0])
            for i in range(nx * ny):
                if i != 2 * nx + 1:
                    self.assertEqual(s[i], 0)

            # Test getting specific states
            self.assertEqual(self.s1.fiber_state(0, 0), sx)
            self.assertEqual(self.s1.fiber_state(1, 2), sm)
            self.assertEqual(self.s1.default_fiber_state(0, 0),
                             self.s1.default_fiber_state(1, 1))

            # Check error messages for set_fiber_state() (shared)
            self.assertRaisesRegex(IndexError, 'x-index out of range',
                                   self.s1.set_fiber_state, sm, -1, 0)
            self.assertRaisesRegex(IndexError, 'x-index out of range',
                                   self.s1.set_fiber_state, sm, nx, 0)
            self.assertRaisesRegex(IndexError, 'y-index out of range',
                                   self.s1.set_fiber_state, sm, 0, -1)
            self.assertRaisesRegex(IndexError, 'y-index out of range',
                                   self.s1.set_fiber_state, sm, 0, ny)
            self.assertRaisesRegex(ValueError, 'both x and y',
                                   self.s1.set_fiber_state, sm, 0)
            self.assertRaisesRegex(ValueError, 'single fiber cell state or',
                                   self.s1.set_fiber_state, sm * 2)

            # Check error messages for fiber_state()
            self.assertRaisesRegex(IndexError, 'x-index out of range',
                                   self.s1.fiber_state, -1, 0)
            self.assertRaisesRegex(IndexError, 'x-index out of range',
                                   self.s1.fiber_state, nx, 0)
            self.assertRaisesRegex(IndexError, 'y-index out of range',
                                   self.s1.fiber_state, 0, -1)
            self.assertRaisesRegex(IndexError, 'y-index out of range',
                                   self.s1.fiber_state, 0, ny)
            self.assertRaisesRegex(ValueError, 'both an x and y',
                                   self.s1.fiber_state, 0)

            # Check error messages for default_fiber_state()
            self.assertRaisesRegex(IndexError, 'x-index out of range',
                                   self.s1.default_fiber_state, -1, 0)
            self.assertRaisesRegex(IndexError, 'x-index out of range',
                                   self.s1.default_fiber_state, nx, 0)
            self.assertRaisesRegex(IndexError, 'y-index out of range',
                                   self.s1.default_fiber_state, 0, -1)
            self.assertRaisesRegex(IndexError, 'y-index out of range',
                                   self.s1.default_fiber_state, 0, ny)
            self.assertRaisesRegex(ValueError, 'both an x and y',
                                   self.s1.default_fiber_state, 0)
        finally:
            self.s1.reset()

    def test_initial_value_expressions(self):
        # Test if initial value expressions are converted to floats
        m = myokit.parse_model('''
            [[model]]
            c.x = 1 + sqrt(3)
            c.y = 1 / c.p
            c.z = 3

            [c]
            t = 0 bind time
            dot(x) = 1 label membrane_potential
                in [mV]
            dot(y) = 2
            dot(z) = 3
            p = log(3)
            q = 0 bind diffusion_current
                in [A/F]
        ''')
        s = myokit.FiberTissueSimulation(
            m, m, ncells_fiber=(1, 1), ncells_tissue=(3, 1))
        x = s.fiber_state()
        self.assertIsInstance(x[0], float)
        self.assertIsInstance(x[1], float)
        self.assertIsInstance(x[2], float)
        self.assertEqual(x, m.initial_values(True))
        self.assertEqual(x, s.default_fiber_state())

        x = s.tissue_state()
        self.assertIsInstance(x[0], float)
        self.assertIsInstance(x[1], float)
        self.assertIsInstance(x[2], float)
        self.assertEqual(x[:3], x[3:6])
        self.assertEqual(x[:3], x[6:])
        self.assertEqual(x, m.initial_values(True) * 3)
        self.assertEqual(x, s.default_tissue_state())

    def test_tissue_state(self):
        # Test the set_tissue_state and set_default_tissue_state methods

        # Check simulation state equals model state
        m = self.mt.count_states()
        nx, ny = self.ntx, self.nty

        sm = self.mt.initial_values(True)
        for i in range(nx):
            for j in range(ny):
                self.assertEqual(sm, self.s1.tissue_state(i, j))

        try:
            # Test setting a full-sized state
            sx = list(range(nx * ny * m))
            self.s1.set_tissue_state(sx)
            self.assertEqual(sx, self.s1.tissue_state())

            # Test setting a single, global state
            sx = [0.0] * m
            self.assertNotEqual(sm, sx)
            self.s1.set_tissue_state(sx)
            for i in range(nx):
                for j in range(ny):
                    self.assertEqual(sx, self.s1.tissue_state(i, j))
            self.assertEqual(sx * (nx * ny), self.s1.tissue_state())
            self.s1.set_tissue_state(sm)
            self.assertEqual(sm * (nx * ny), self.s1.tissue_state())

            # Test setting the state of a single cell
            x, y = 1, 2
            self.s1.set_tissue_state(sx, x, y)
            for i in range(nx):
                for j in range(ny):
                    if i == x and j == y:
                        self.assertEqual(self.s1.tissue_state(i, j), sx)
                    else:
                        self.assertEqual(self.s1.tissue_state(i, j), sm)

            # Test indexing in state vector is x first, then y
            self.s1.set_tissue_state(sx)
            self.s1.set_tissue_state(sm, x=1, y=2)
            s = self.s1.tissue_state()[::m]
            self.assertEqual(s[2 * nx + 1], sm[0])
            for i in range(nx * ny):
                if i != 2 * nx + 1:
                    self.assertEqual(s[i], 0)

            # Test getting specific states
            self.assertEqual(self.s1.tissue_state(0, 0), sx)
            self.assertEqual(self.s1.tissue_state(1, 2), sm)
            self.assertEqual(self.s1.default_tissue_state(0, 0),
                             self.s1.default_tissue_state(1, 1))

            # Check error messages for set_tissue_state() (shared)
            self.assertRaisesRegex(IndexError, 'x-index out of range',
                                   self.s1.set_tissue_state, sm, -1, 0)
            self.assertRaisesRegex(IndexError, 'x-index out of range',
                                   self.s1.set_tissue_state, sm, nx, 0)
            self.assertRaisesRegex(IndexError, 'y-index out of range',
                                   self.s1.set_tissue_state, sm, 0, -1)
            self.assertRaisesRegex(IndexError, 'y-index out of range',
                                   self.s1.set_tissue_state, sm, 0, ny)
            self.assertRaisesRegex(ValueError, 'both x and y',
                                   self.s1.set_tissue_state, sm, 0)
            self.assertRaisesRegex(ValueError, 'single tissue cell state or',
                                   self.s1.set_tissue_state, sm * 2)

            # Check error messages for tissue_state()
            self.assertRaisesRegex(IndexError, 'x-index out of range',
                                   self.s1.tissue_state, -1, 0)
            self.assertRaisesRegex(IndexError, 'x-index out of range',
                                   self.s1.tissue_state, nx, 0)
            self.assertRaisesRegex(IndexError, 'y-index out of range',
                                   self.s1.tissue_state, 0, -1)
            self.assertRaisesRegex(IndexError, 'y-index out of range',
                                   self.s1.tissue_state, 0, ny)
            self.assertRaisesRegex(ValueError, 'both an x and y',
                                   self.s1.tissue_state, 0)

            # Check error messages for default_tissue_state()
            self.assertRaisesRegex(IndexError, 'x-index out of range',
                                   self.s1.default_tissue_state, -1, 0)
            self.assertRaisesRegex(IndexError, 'x-index out of range',
                                   self.s1.default_tissue_state, nx, 0)
            self.assertRaisesRegex(IndexError, 'y-index out of range',
                                   self.s1.default_tissue_state, 0, -1)
            self.assertRaisesRegex(IndexError, 'y-index out of range',
                                   self.s1.default_tissue_state, 0, ny)
            self.assertRaisesRegex(ValueError, 'both an x and y',
                                   self.s1.default_tissue_state, 0)
        finally:
            self.s1.reset()

    def test_step_size(self):
        # Tests setting/getting step size
        try:
            self.assertEqual(self.s1.step_size(), 0.0012)
            self.s1.set_step_size(0.0011)
            self.assertEqual(self.s1.step_size(), 0.0011)

            self.assertRaisesRegex(ValueError, 'greater than zero',
                                   self.s1.set_step_size, -1)
            self.assertRaisesRegex(ValueError, 'greater than zero',
                                   self.s1.set_step_size, 0)
        finally:
            self.s1.set_step_size(0.0012)


@unittest.skipIf(not OpenCL_FOUND, 'OpenCL not found on this system.')
class FiberTissueSimulationFindNanTest(unittest.TestCase):
    """ Tests the FiberTissue simulation's find_nan method. """

    @classmethod
    def setUpClass(cls):
        # Load models and protocols for use in testing

        # Fiber model, tissue model, protocol
        cls.mf = myokit.load_model(
            os.path.join(DIR_DATA, 'dn-1985-normalised.mmt'))
        cls.mt = myokit.load_model(os.path.join(DIR_DATA, 'lr-1991.mmt'))

        # Fiber/Tissue sizes (small sim)
        cls.nfx = 8
        cls.nfy = 4
        cls.ntx = 8
        cls.nty = 6

        # Standard protocols
        cls.p1 = myokit.pacing.blocktrain(1000, 2, offset=.01)

        # Protocol with error at t = 2
        cls.p2 = myokit.Protocol()
        cls.p2.schedule(period=0, duration=5, level=5e3, start=2)

        # Protocol with error at t = 1.234 (in tissue?)
        cls.p3 = myokit.Protocol()
        cls.p3.schedule(period=0, duration=5, level=-1, start=1.234)
        v = cls.mt.get('ina.m')
        v.set_rhs('if(membrane.V < -100, 1e4, ' + v.rhs().code() + ')')

        # Shared simulations
        cls._s1 = None

    @property
    def s1(self):
        if self._s1 is None:
            self._s1 = myokit.FiberTissueSimulation(
                self.mf,
                self.mt,
                self.p1,
                ncells_fiber=(self.nfx, self.nfy),
                ncells_tissue=(self.ntx, self.nty),
                nx_paced=10,
                g_fiber=(235, 100),
                g_tissue=(9, 5),
                g_fiber_tissue=9,
                dt=0.0012,
            )
        return self._s1

    def test_big_stimulus(self):
        # Tests if NaNs are detected after a massive stimulus

        try:
            self.s1.set_protocol(self.p2)

            # Automatic detection
            with WarningCollector():
                self.assertRaisesRegex(
                    myokit.SimulationError,
                    'numerical error in fiber simulation at t = 2.0',
                    self.s1.run, 10, log_interval=1)

            # Automatic detection, but not enough information
            self.s1.reset()
            with WarningCollector():
                self.assertRaisesRegex(
                    myokit.SimulationError, 'Unable to pinpoint',
                    self.s1.run, 5, logf=['membrane.V'], logt=myokit.LOG_NONE,
                    log_interval=0.1)

            # Offline detection
            # res = part, time, icell, var, value, states, bound
            self.s1.reset()
            df, dt = self.s1.run(5, log_interval=0.1, report_nan=False)
            res = self.s1.find_nan(df, dt)
            self.assertEqual(round(res[1], 1), 2.0)

            # Missing state and bound var in fiber log
            d2 = df.clone()
            del d2['membrane.V', 0, 0]
            self.assertRaisesRegex(
                myokit.FindNanError, 'fiber(.+)membrane.V',
                self.s1.find_nan, d2, dt)
            d2 = df.clone()
            del d2['engine.time']
            self.assertRaisesRegex(
                myokit.FindNanError, 'fiber(.+)engine.time',
                self.s1.find_nan, d2, dt)

            # Missing state and bound var in tissue log
            d2 = dt.clone()
            del d2['membrane.V', 0, 0]
            self.assertRaisesRegex(
                myokit.FindNanError, 'tissue(.+)membrane.V',
                self.s1.find_nan, df, d2)
            d2 = dt.clone()
            del d2['engine.time']
            self.assertRaisesRegex(
                myokit.FindNanError, 'tissue(.+)engine.time',
                self.s1.find_nan, df, d2)

            # NaN at first data point
            d2 = df.clone()
            d2['membrane.V', 0, 0][0] = float('nan')
            self.assertRaisesRegex(
                myokit.FindNanError, 'first data point',
                self.s1.find_nan, d2, dt)

            # No NaN
            d2 = df.trim_right(1.9)
            d3 = dt.trim_right(1.9)
            self.assertRaisesRegex(
                myokit.FindNanError, 'Error condition not found',
                self.s1.find_nan, d2, d3)

        finally:
            self.s1.set_protocol(self.p1)
            self.s1.reset()

    def test_neg_stimulus(self):
        # Tests if NaNs are detected in the tissue, after a negative stimulus

        try:
            self.s1.set_protocol(self.p3)
            with WarningCollector():
                self.assertRaisesRegex(
                    myokit.SimulationError, 'in tissue simulation at t = 1.7',
                    self.s1.run, 5)
        finally:
            self.s1.set_protocol(self.p1)
            self.s1.reset()

    def test_big_step(self):
        # Tests if NaNs are detected in the diffusion current

        dt = self.s1.step_size()
        try:
            self.s1.set_step_size(0.01)
            with WarningCollector():
                self.assertRaisesRegex(
                    myokit.SimulationError, 'when membrane.i_diff',
                    self.s1.run, 5)
        finally:
            self.s1.set_step_size(dt)
            self.s1.reset()


if __name__ == '__main__':

    import sys
    if '-v' in sys.argv:
        print('Running in debug/verbose mode')
        debug = True
    else:
        print('Add -v for more debug output')

    import warnings
    warnings.simplefilter('always')

    unittest.main()
